/* ================================================================
 * ================================================================
 *
 * A python module that is compiled and provides an interface between
 * C++ classes for heavy-weight data analysis and python scripts
 * that request and plot the data.
 *
 * Contains a static instance of DataProvider, ensuring data persistence
 * once loaded from a file
 *
 * Author: L. Pitonakova (http://lenkaspace.net)
 * License: GNU General Public License. Please credit me when using my work.
 *
 * ================================================================
 * ================================================================
 */


#include <Python.h>
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <vector>

#include "helpers/pythonFormatHelper.h"
#include "helpers/ioHelper.h"
#include "helpers/dataProvider.h"
#include <event.h>


static DataProvider dataProvider;



/* =====================================================================
                    SINGLE VALUE METRICS - NOVELTY
   ===================================================================== */

/**
 * getNetworkTrainingTime(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,startTime_,endTime_)
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 *
 * @return 1D array of network training times, i.e., how long it took a network not to output novelty
 *    0th dimension is values, includes those for different runs and different robots
 */
static PyObject* gaData_getNetworkTrainingTime(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, startTime_, endTime_;
   float noveltyValueThreshold_;

   if (!PyArg_ParseTuple(args, "s|i|i|f|i|i", &dataPath_,&numOfRuns_,&numOfRobots_, &noveltyValueThreshold_, &startTime_, &endTime_)) return NULL;
   printf("Getting network training time for %s\n   (num of runs: %d, num of robots: %d,  novelty value threshold: %f  time span: %d-%d  )\n ", dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, startTime_, endTime_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNetworkTrainingTime(dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, startTime_, endTime_));
}

/**
 * getNumOfRunsNovelObjectReported(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,noveltyDetectionRange_,eventType_)
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 *
 * @return 1D array of number of runs during which a novelty object was reported in pass N while any robot was within its range
 *    0th dimension is passes
 */
static PyObject* gaData_getNumOfRunsNovelObjectReported(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, eventType_;
   float noveltyValueThreshold_, noveltyDetectionRange_;

   if (!PyArg_ParseTuple(args, "s|i|i|f|f|i", &dataPath_,&numOfRuns_,&numOfRobots_, &noveltyValueThreshold_, &noveltyDetectionRange_, &eventType_)) return NULL;
   printf("Getting num of runs novel object reported for %s\n   (num of runs: %d, num of robots: %d,\n   novelty value threshold: %f, novelty detection range: %f,  event type: %d  )\n ", dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, noveltyDetectionRange_, eventType_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfRunsNovelObjectReported(dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, noveltyDetectionRange_, Event::TYPE(eventType_)));
}


/**
 * getNumOfNoveltyTruePositives(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,noveltyDetectionRange_,eventType_)
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 *
 * @return 2D array of number of true positives in a certain pass and run, across all robots.
 * A true positive is when a robot that is near a novel object reports a novelty.
 *    0th dimension is passes
 *    1st dimension is runs
 */
static PyObject* gaData_getNumOfNoveltyTruePositives(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, eventType_;
   float noveltyValueThreshold_, noveltyDetectionRange_;

   if (!PyArg_ParseTuple(args, "s|i|i|f|f|i", &dataPath_,&numOfRuns_,&numOfRobots_, &noveltyValueThreshold_, &noveltyDetectionRange_, &eventType_)) return NULL;
   printf("Getting novelty true positives for %s\n   (num of runs: %d, num of robots: %d,\n   novelty value threshold: %f  novelty detection range: %f,  event type: %d  )\n ", dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, noveltyDetectionRange_, eventType_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfNoveltyTruePositives(dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, noveltyDetectionRange_, Event::TYPE(eventType_)));
}


/**
 * getNumOfNoveltyFalseNegatives(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,noveltyDetectionRange_,eventType_)
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 *
 * @return 2D array of number of false negatives in a certain pass and run, across all robots.
 * A false negative is when a robot that is near a novel object does not report a novelty.
 *    0th dimension is passes
 *    1st dimension is runs
 */
static PyObject* gaData_getNumOfNoveltyFalseNegatives(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, eventType_;
   float noveltyValueThreshold_, noveltyDetectionRange_;

   if (!PyArg_ParseTuple(args, "s|i|i|f|f|i", &dataPath_,&numOfRuns_,&numOfRobots_, &noveltyValueThreshold_, &noveltyDetectionRange_, &eventType_)) return NULL;
   printf("Getting novelty false negatives for %s\n   (num of runs: %d, num of robots: %d,\n   novelty value threshold: %f  novelty detection range: %f,  event type: %d  )\n ", dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, noveltyDetectionRange_, eventType_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfNoveltyFalseNegatives(dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, noveltyDetectionRange_, Event::TYPE(eventType_)));
}

/**
 * getNumOfNoveltyFalsePositives(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,noveltyDetectionRange_,eventType_)
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 *
 * @return 2D array of number of false positives in a certain pass and run, across all robots.
 * A false positive is when a robot that is not near a novel object reports a novelty
 *    0th dimension is passes
 *    1st dimension is runs
 */
static PyObject* gaData_getNumOfNoveltyFalsePositives(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, eventType_;
   float noveltyValueThreshold_, noveltyDetectionRange_;

   if (!PyArg_ParseTuple(args, "s|i|i|f|f|i", &dataPath_,&numOfRuns_,&numOfRobots_, &noveltyValueThreshold_, &noveltyDetectionRange_, &eventType_)) return NULL;
   printf("Getting novelty false positives for %s\n   (num of runs: %d, num of robots: %d,\n   novelty value threshold: %f  novelty detection range: %f,  event type: %d  )\n ", dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, noveltyDetectionRange_, eventType_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfNoveltyFalsePositives(dataPath_, numOfRuns_, numOfRobots_, noveltyValueThreshold_, noveltyDetectionRange_, Event::TYPE(eventType_)));
}





/* =====================================================================
                       TIME SERIES METRICS - NOVELTY
   ===================================================================== */

/**
 * getNoveltyValues(dataPath_,numOfRuns_,numOfRobots_)
 * @return 3D array of novelty values for a specific experiment, where:
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
static PyObject* gaData_getNoveltyValues(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting novelty data for %s\n   (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNoveltyValuesOverTime(dataPath_, numOfRuns_, numOfRobots_));
}

/**
 * getWinningNeuronErrorQuantizationOverTime(dataPath_,numOfRuns_,numOfRobots_)
 * @return 3D array of winning neuron quantization error (distance from current input) for a specific experiment, where:
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
static PyObject* gaData_getWinningNeuronQuantizationErrorOverTime(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting winning neuron quantization errors for %s\n   (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetWinningNeuronQuantizationErrorOverTime(dataPath_, numOfRuns_, numOfRobots_));
}

/**
 * gaData_getWinningNeuronNeigboursErrorOverTime(dataPath_,numOfRuns_,numOfRobots_)
 * @return 3D array of winning neuron's neigbours quantization error (distance from current input) for a specific experiment, where:
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
static PyObject* gaData_getWinningNeuronNeigboursQuantizationErrorOverTime(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting winning neuron's neighbour errors for %s\n  (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetWinningNeuronNeigboursQuantizationErrorOverTime(dataPath_, numOfRuns_, numOfRobots_));
}


/**
 * getRobotsAroundNoveltyTimes(dataPath_,numOfRuns_,numOfRobots_,noveltyDetectionRange_)
 * @return 3D array of start and end times where robots were around a novelty that has occurred, where:
 *    0th dimension is robot id
 *    1st dimension is run number
 *    2nd dimension is individual values, listed as startTime,endTime,startTime,endTime,...
 */
static PyObject* gaData_getRobotsAroundNoveltyTimes(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   float noveltyDetectionRange_;
   if (!PyArg_ParseTuple(args, "s|i|i|f", &dataPath_,&numOfRuns_,&numOfRobots_, &noveltyDetectionRange_)) return NULL;
   printf("Getting robots around env changes data for %s\n   (num of runs: %d, num of robots: %d  novelty detection range: %f)\n ", dataPath_, numOfRuns_, numOfRobots_, noveltyDetectionRange_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetRobotsAroundNoveltyTimes(dataPath_, numOfRuns_, numOfRobots_, noveltyDetectionRange_, MIN_DELAY_AFTER_ENV_CHANGE));
}


/**
 * getNoveltyValuesByDistanceToNovelty(dataPath_,numOfRuns_,numOfRobots_,numOfPasses_, noveltyDetectionRange_)
 *
 * :param numOfPasses_: how many passes after environment changed to include
 * :param noveltyDetectionRange_: how far from novelty should robot sensors be able to pick it up
 *
 * @return 4D array of novelty values organised by distance from a novelty, where:
 *    0th dimension is robot id
 *    1st dimension is pass number
 *    2nd dimension is run number
 *    3rd dimension is individual values, listed by distance from novelty, in range [-noveltyDetectionRange_; noveltyDetectionRange_] (so middle values are for when robot at novelty)
 */
static PyObject* gaData_getNoveltyValuesByDistanceToNovelty(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, numOfPasses_;
   float noveltyDetectionRange_;

   if (!PyArg_ParseTuple(args, "s|i|i|i|f", &dataPath_,&numOfRuns_,&numOfRobots_, &numOfPasses_, &noveltyDetectionRange_)) return NULL;
   printf("Getting novelty data by distance to env change for %s\n   (num of runs: %d, num of robots: %d,  num of passes: %d, novelty detection range: %f)\n ", dataPath_, numOfRuns_, numOfRobots_, numOfPasses_, noveltyDetectionRange_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNoveltyValuesByDistanceToNovelty(dataPath_, numOfRuns_, numOfRobots_, numOfPasses_, noveltyDetectionRange_));
}




/* =====================================================================
                       TIME SERIES METRICS - EVENTS
   ===================================================================== */

/**
 * getRobotTeleportTimes(dataPath_,numOfRuns_,numOfRobots_)
 * @return 3D array of times at which robots reached an instant waypoint where:
 *    0th dimension is robot id
 *    1st dimension is run number
 *    2nd dimension has the teleport times
 */
static PyObject* gaData_getRobotTeleportTimes(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting robot teleport times data for %s\n   (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetRobotTeleportTimes(dataPath_, numOfRuns_, numOfRobots_));
}

/**
 * gaData_getRobotReachedWaypointTimes(dataPath_,numOfRuns_,numOfRobots_,waypointId_)
 * @return 3D array of times at which robots reached a wapying of a certain id(string), where_
 *    0th dimension is robot id
 *    1st dimension is run number
 *    2nd dimension has the waypoint reached times
 */
static PyObject* gaData_getRobotReachedWaypointTimes(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   const char* waypointId_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i|s", &dataPath_,&numOfRuns_,&numOfRobots_, &waypointId_)) return NULL;
   printf("Getting robot reached waypoint %s times for %s\n   (num of runs: %d, num of robots: %d)\n ", waypointId_, dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetRobotReachedWaypointTimes(dataPath_, numOfRuns_, numOfRobots_, waypointId_));
}

/**
 * getEventTimes(dataPath_,numOfRuns_,numOfRobots_,eventType_)
 * @return 2D array of times at which an event of a certain ID occurred where:
 *    0th dimension is run number
 *    1st dimension has the event times
 */
static PyObject* gaData_getEventTimes(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int eventType_, numOfRuns_, numOfRobots_;
   if (!PyArg_ParseTuple(args, "s|i|i|i", &dataPath_,&numOfRuns_,&numOfRobots_,&eventType_)) return NULL;
   printf("Getting event times data for %s\n   (num of runs: %d, num of robots: %d, event id: %d)\n ", dataPath_, numOfRuns_, numOfRobots_, eventType_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetEventTimes(dataPath_, numOfRuns_, numOfRobots_, Event::TYPE(eventType_)));
}





/* =====================================================================
                    TIME SERIES METRICS - NEURAL NETWORK
   ===================================================================== */

/**
 * getNumOfNeuronsOverTime(dataPath_,numOfRuns_,numOfRobots_)
 * @return 3D array of number of neurons for a specific experiment, where:
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
static PyObject* gaData_getNumOfNeuronsOverTime(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting number of neurons over time for %s\n   (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfNeuronsOverTime(dataPath_, numOfRuns_, numOfRobots_));
}

/**
 * getNumOfUsedNeuronsOverTimeSince(dataPath_,numOfRuns_,numOfRobots_,sinceTime_)
 * @return 3D array of how many neurons have been used (fired as winning) in each time step since time step N.
 * E.g. if there is total of 3 neurons and winning neurons for 5 time steps are 1,2,3,2,3, Then, for time step 4 (winning neuron 2), 3 neurons have been used since step 0, and 2 neurons since step 3.
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
static PyObject* gaData_getNumOfUsedNeuronsOverTimeSince(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, sinceTime_;
   if (!PyArg_ParseTuple(args, "s|i|i|i", &dataPath_,&numOfRuns_,&numOfRobots_,&sinceTime_)) return NULL;
   printf("Getting number of used neurons over time for %s\n   (num of runs: %d, num of robots: %d, since %d)\n ", dataPath_, numOfRuns_, numOfRobots_, sinceTime_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfUsedNeuronsOverTimeSince(dataPath_, numOfRuns_, numOfRobots_, sinceTime_));
}



/* =====================================================================
                    SINGLE VALUE METRICS - NEURAL NETWORK
   ===================================================================== */

/**
 * getNumOfNeurons(dataPath_,numOfRuns_,numOfRobots_)
 * @return 1D array of maximum number of neurons, where
 *    0th dimension is values, includes those for different runs and different robots
 */
static PyObject* gaData_getNumOfNeurons(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting number of neurons for %s\n   (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfNeurons(dataPath_, numOfRuns_, numOfRobots_));
}

/**
 * gaData_getNumOfInWeights(dataPath_,numOfRuns_,numOfRobots_)
 * @return 1D array of average total number of neuron in weights (averaged for each robot in each run), where
 *    0th dimension is values, includes those for different runs and different robots
 */
static PyObject* gaData_getNumOfInWeights(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting number of in weights for %s\n   (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfInWeights(dataPath_, numOfRuns_, numOfRobots_));
}

/**
 * gaData_getNumOfInWeightsPerNeuron(dataPath_,numOfRuns_,numOfRobots_)
 * @return 1D array of average number of neuron in weights per neuron (averaged for each robot in each run), where
 *    0th dimension is values, includes those for different runs and different robots
 */
static PyObject* gaData_getNumOfInWeightsPerNeuron(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting number of in weights for %s\n   (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfInWeightsPerNeuron(dataPath_, numOfRuns_, numOfRobots_));
}

/**
 * gaData_getNumOfConnections(dataPath_,numOfRuns_,numOfRobots_)
 * @return 1D array of average total number of inter-neuron connections (averaged for each robot in each run), where
 *    0th dimension is values, includes those for different runs and different robots
 */
static PyObject* gaData_getNumOfConnections(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_;
   if (!PyArg_ParseTuple(args, "s|i|i", &dataPath_,&numOfRuns_,&numOfRobots_)) return NULL;
   printf("Getting number of connections for %s\n   (num of runs: %d, num of robots: %d)\n ", dataPath_, numOfRuns_, numOfRobots_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNumOfConnections(dataPath_, numOfRuns_, numOfRobots_));
}

/**
 * getNetworkGrowthRate(dataPath_,numOfRuns_,numOfRobots_,maxTime_)
 * @return return 1D array of average growth rate (% size increase) per second where
 *    0th dimension is values, includes those for different runs and different robots
 */
static PyObject* gaData_getNetworkGrowthRate(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, maxTime_;
   if (!PyArg_ParseTuple(args, "s|i|i|i", &dataPath_,&numOfRuns_,&numOfRobots_,&maxTime_)) return NULL;
   printf("Getting network growth rate for %s\n   (num of runs: %d, num of robots: %d maxTime: %d)\n ", dataPath_, numOfRuns_, numOfRobots_, maxTime_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNetworkGrowthRate(dataPath_, numOfRuns_, numOfRobots_, maxTime_));
}

/* =====================================================================
                       HISTOGRAM METRICS - NOVELTY
   ===================================================================== */

/**
 * getNeuronUsageAsBins(dataPath_,numOfRuns_,numOfRobots_,startTime_,endTime_)
 *
 * maxFiredCount_ defines the last bin, all counts where neuron fired more than maxFiredTimes_ will be added into the last bin.
 *
 * @return 2D array of bins for how many times a neuron was used where:
 *    0th dimension is robot id
 *    1st dimension is number of times used, with values representing percentage of neurons out of final network size, averaged over all runs
 *
 * Note: This function only works correctly for network that doesn't decrease in size!
 * Note: for each robot, the size of the array of counts may be different.
 */
static PyObject* gaData_getNeuronUsageAsBins(PyObject *self, PyObject *args) {
   //-- parse arguments
   const char* dataPath_;
   int numOfRobots_, numOfRuns_, startTime_, endTime_;
   if (!PyArg_ParseTuple(args, "s|i|i|i|i|i", &dataPath_,&numOfRuns_,&numOfRobots_,&startTime_, &endTime_)) return NULL;
   printf("Getting neuron usage in bins for %s\n   (num of runs: %d, num of robots: %d,  time span: %d-%d  )\n ", dataPath_, numOfRuns_, numOfRobots_, startTime_, endTime_);
   return PythonFormatHelper::PythonListFromVector(dataProvider.GetNeuronUsageAsBins(dataPath_, numOfRuns_, numOfRobots_, startTime_, endTime_));
}



/* =====================================================================
                               MODULE DEFINITIONS
   ===================================================================== */

/**
 * Method definition object for this extension. The arguments mean:
 *    ml_name: The name of the method
 *    ml_meth: Function pointer to the method implementation
 *    ml_flags: Flags indicating special features of this method, such as accepting arguments, accepting keyword arguments, being a class method, or being a static method of a class.
 *    ml_doc:  Contents of this method's docstring
 */
static PyMethodDef gaData_moduleMethods[] = {
   {
      "getNoveltyValues", gaData_getNoveltyValues,
      METH_VARARGS,
      "getNoveltyValues(dataPath_,numOfRuns_,numOfRobots_). Return 3D array of novelty values for a specific experiment."
   },
   {
      "getRobotTeleportTimes", gaData_getRobotTeleportTimes,
      METH_VARARGS,
      "getRobotTeleportTimes(dataPath_,numOfRuns_,numOfRobots_). Return 3D array of times at which robots reached an instant waypoint."
   },
   {
      "getRobotReachedWaypointTimes", gaData_getRobotReachedWaypointTimes,
      METH_VARARGS,
      "gaData_getRobotReachedWaypointTimes(dataPath_,numOfRuns_,numOfRobots_,waypointId_) \
      3D array of times at which robots reached a wapying of a certain id(string), where_ \
      0th dimension is robot id, 1st dimension is run number, 2nd dimension has the waypoint reached times"
   },
   {
      "getEventTimes", gaData_getEventTimes,
      METH_VARARGS,
      "getEventTimes(dataPath_,numOfRuns_,numOfRobots_,eventType_). Return 2D array of times at which an event of a certain ID occurred where: 0th dimension is run number, 1st dimension has the event times"
   },
   {
      "getRobotsAroundNoveltyTimes", gaData_getRobotsAroundNoveltyTimes,
      METH_VARARGS,
      "getRobotsAroundNoveltyTimes(dataPath_,numOfRuns_,numOfRobots_,noveltyDetectionRange_). Return 3D array of start and end times where robots were around a novelty that has occurred, where: \
      0th dimension is robot id, 1st dimension is run number, 2nd dimension is individual values, listed as startTime,endTime,startTime,endTime,.."
   },
   {
      "getNoveltyValuesByDistanceToNovelty", gaData_getNoveltyValuesByDistanceToNovelty,
      METH_VARARGS,
      "getNoveltyValuesByDistanceToNovelty(dataPath_,numOfRuns_,numOfRobots_,numOfPasses_, noveltyDetectionRange_). Return 4D array of novelty values organised by distance from a novelty, where: \
       0th dimension is robot id, 1st dimension is pass number, 2nd dimension is run number, 3rd dimension is individual values, listed by distance from novelty, in range [-noveltyDetectionRange_; \
      noveltyDetectionRange_] (so middle values are for when robot at novelty)"
   },

   {
      "getNetworkTrainingTime", gaData_getNetworkTrainingTime,
      METH_VARARGS,
      "getNetworkTrainingTime(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,startTime_,endTime_). \
      Return 1D array of network training times, i.e., how long it took a network not to output novelty, where: \
      0th dimension is values, includes those for different runs and different robots"
   },
   {
      "getNumOfRunsNovelObjectReported", gaData_getNumOfRunsNovelObjectReported,
      METH_VARARGS,
      "getNumOfRunsNovelObjectReported(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,noveltyDetectionRange_,eventType_). Return 1D array of number of runs during which a novelty \
      object was reported in pass N while any robot was within its range where 0th dimension is passes"
   },
   {
      "getNumOfNoveltyTruePositives", gaData_getNumOfNoveltyTruePositives,
      METH_VARARGS,
      "getNumOfNoveltyTruePositives(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,noveltyDetectionRange_,eventType_). \
      Return 2D array of number of true positives in a certain pass and run, across all robots. \
      A true positive is when a robot that is near a novel object reports a novelty. \
      0th dimension is passes, 1st dimension is runs"
   },
   {
      "getNumOfNoveltyFalseNegatives", gaData_getNumOfNoveltyFalseNegatives,
      METH_VARARGS,
      "getNumOfNoveltyFalseNegatives(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,noveltyDetectionRange_,eventType_). \
      Return 2D array of number of false negatives in a certain pass and run, across all robots. \
      A false negative is when a robot that is near a novel object does not report a novelty. \
      0th dimension is passes, 1st dimension is runs"
   },
   {
      "getNumOfNoveltyFalsePositives", gaData_getNumOfNoveltyFalsePositives,
      METH_VARARGS,
      "getNumOfNoveltyFalsePositives(dataPath_,numOfRuns_,numOfRobots_,noveltyValueThreshold_,noveltyDetectionRange_,eventType_). \
      Return 2D array of number of false positives in a certain pass and run, across all robots. \
      A false positive is when a robot that is not near a novel object reports a novelty. \
      0th dimension is passes, 1st dimension is runs"
   },
   {
      "getNumOfNeuronsOverTime", gaData_getNumOfNeuronsOverTime,
      METH_VARARGS,
      "getNumOfNeuronsOverTime(dataPath_,numOfRuns_,numOfRobots_). \
      Return 3D array of number of neurons for a specific experiment, where: \
      0th dimension is robot id, 1st dimension is time, 2nd dimension is run number"
   },
   {
      "getNumOfUsedNeuronsOverTimeSince", gaData_getNumOfUsedNeuronsOverTimeSince,
      METH_VARARGS,
      "getNumOfUsedNeuronsOverTimeSince(dataPath_,numOfRuns_,numOfRobots_,sinceTime_) \
      Return 3D array of how many neurons have been used (fired as winning) in each time step since time step N. \
      E.g. if there is total of 3 neurons and winning neurons for 5 time steps are 1,2,3,2,3, Then, \
      for time step 4 (winning neuron 2), 3 neurons have been used since step 0, and 2 neurons since step 3. \
      0th dimension is robot id, 1st dimension is time, 2nd dimension is run number"
   },
   {
      "getNumOfNeurons", gaData_getNumOfNeurons,
      METH_VARARGS,
      "getNumOfNeurons(dataPath_,numOfRuns_,numOfRobots_) \
      Return 1D array of maximum number of neurons, where \
      0th dimension is values, includes those for different runs and different robots"
   },
   {
      "getNumOfInWeights", gaData_getNumOfInWeights,
      METH_VARARGS,
      "getNumOfInWeights(dataPath_,numOfRuns_,numOfRobots_) \
      Return 1D array of average total number of neuron in weights (averaged for each robot in each run), where \
      0th dimension is values, includes those for different runs and different robots"
   },
   {
      "getNumOfInWeightsPerNeuron", gaData_getNumOfInWeightsPerNeuron,
      METH_VARARGS,
      "getNumOfInWeightsPerNeuron(dataPath_,numOfRuns_,numOfRobots_) \
      Return 1D array of average number of neuron in weights per neuron (averaged for each robot in each run), where \
      0th dimension is values, includes those for different runs and different robots"
   },
   {
      "getNumOfConnections", gaData_getNumOfConnections,
      METH_VARARGS,
      "getNumOfConnections(dataPath_,numOfRuns_,numOfRobots_) \
      Return 1D array of average total number of inter-neuron connections (averaged for each robot in each run), where \
      0th dimension is values, includes those for different runs and different robots"
   },
   {
      "getNetworkGrowthRate", gaData_getNetworkGrowthRate,
      METH_VARARGS,
      "getNetworkGrowthRate(dataPath_,numOfRuns_,numOfRobots_,maxTime_) \
      Return 1D array of average growth rate (% size increase) per second where \
      0th dimension is values, includes those for different runs and different robots"
   },
   {
      "getNeuronUsageAsBins", gaData_getNeuronUsageAsBins,
      METH_VARARGS,
      "getNeuronUsageAsBins(dataPath_,numOfRuns_,numOfRobots_,startTime_,endTime_) \
      Return 2D array of bins for how many times a neuron was used where: \
      0th dimension is robot id, 1st dimension is number of times used, with values representing percentage of neurons out of final network size.\
      Note: This function only works correctly for network that doesn't decrease in size! \
      Note: for each robot, the size of the array of counts may be different."
   },
   {
      "getWinningNeuronQuantizationErrorOverTime", gaData_getWinningNeuronQuantizationErrorOverTime,
      METH_VARARGS,
      "getWinningNeuronQuantizationErrorOverTime(dataPath_,numOfRuns_,numOfRobots_) \
      Return 3D array of winning neuron quantization error (distance from current input) for a specific experiment, where: \
      0th dimension is robot id, 1st dimension is time, 2nd dimension is run number."
   },
   {
      "getWinningNeuronNeigboursQuantizationErrorOverTime", gaData_getWinningNeuronNeigboursQuantizationErrorOverTime,
      METH_VARARGS,
      "getWinningNeuronNeigboursQuantizationErrorOverTime(dataPath_,numOfRuns_,numOfRobots_) \
      Return 3D array of winning neuron's neigbours quantization error (distance from current input) for a specific experiment, where: \
      0th dimension is robot id, 1st dimension is time, 2nd dimension is run number."
   },

   {NULL, NULL, 0, NULL}
};

/**
 * Module definition
 */
static struct PyModuleDef gaData_moduleDefinition = {
    PyModuleDef_HEAD_INIT,
    "gaData_module",
    "A Python module that provides data from the gwrnn experiments",
    -1, 
    gaData_moduleMethods
};

/**
 * Module initialization
 * It is important that this function is named PyInit_[[module_name]] exactly, and matches the name keyword argument in setup.py's setup() call.
 */
PyMODINIT_FUNC PyInit_gaData_module(void) {
    Py_Initialize();
    return PyModule_Create(&gaData_moduleDefinition);
};
