/* ================================================================
 * ================================================================
 *
 * The main data provider class, called from gaData_module.
 *
 * Author: L. Pitonakova (http://lenkaspace.net)
 * License: GNU General Public License. Please credit me when using my work.
 *
 * ================================================================
 * ================================================================
 */
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <numeric>
#include <algorithm>
#include <cmath>

#include "dataProvider.h"
#include "ioHelper.h"
#include "mathHelper.h"

DataProvider::DataProvider() {

}

DataProvider::~DataProvider() {

}


/* =====================================================================
                    SINGLE VALUE METRICS - NOVELTY
   ===================================================================== */


/**
 * Return 1D array of network training times, i.e., how long it took a network not to output novelty
 *    0th dimension is values, includes those for different runs and different robots
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 */
std::vector<float> DataProvider::GetNetworkTrainingTime(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, int startTime_, int endTime_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<float> returnArray;

   int lastTimeThresholdFound;
   for (int robotId = 0; robotId < (int)(robotNoveltyValues.size()); robotId++) {
      for (int runNo = 0; runNo < (int)(robotNoveltyValues[robotId][0].size()); runNo++) {
         lastTimeThresholdFound = endTime_;
         for (int t = std::max(0, startTime_); t < std::min(endTime_,(int)(robotNoveltyValues[robotId].size())); t++) {
            if (robotNoveltyValues[robotId][t][runNo] >= noveltyValueThreshold_) {
               lastTimeThresholdFound = t;
               //printf("robotId= %d runNo = %d t=%d noveltyVal=%f lastTime %f found: %d\n",robotId, runNo, t, robotNoveltyValues[robotId][t][runNo], noveltyValueThreshold_, lastTimeThresholdFound);

            }
         }
         //printf("--- robotId= %d runNo = %d adding val %d \n", robotId, runNo, lastTimeThresholdFound);
         returnArray.push_back(lastTimeThresholdFound);
      }
   }

   return returnArray;
}

/**
 * Return 1D array of number of runs during which a novelty object was reported in pass N while any robot was within its range
 *    0th dimension is passes
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 */
std::vector<float> DataProvider::GetNumOfRunsNovelObjectReported(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, float noveltyDetectionRange_, Event::TYPE eventType_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_, MAX_NUM_OF_PASSES, noveltyValueThreshold_, noveltyDetectionRange_) != lastCalculationSetIdentifier_processNoveltyValues) { ProcessNoveltyValues(dataPath_, numOfRuns_, numOfRobots_, MAX_NUM_OF_PASSES, noveltyValueThreshold_, noveltyDetectionRange_); }

   int maxNumOfPasses = MAX_NUM_OF_PASSES;
   std::vector<float> returnArray (maxNumOfPasses, 0);
   for (int robotId = 0; robotId < noveltyValuesByDistanceToNovelty.size(); robotId++) {
      for (int passNo = 0; passNo < (int)(noveltyValuesByDistanceToNovelty[robotId].size()); passNo ++) {
         for (int runNo = 0; runNo < (int)(noveltyValuesByDistanceToNovelty[robotId][passNo].size()); runNo ++) {
            for (int i=0; i < (int)(noveltyValuesByDistanceToNovelty[robotId][passNo][runNo].size()); i ++) {
               //printf("%f ", noveltyValsByDistance[robotId][passNo][runNo][i]);
               if (noveltyValuesByDistanceToNovelty[robotId][passNo][runNo][i] >= noveltyValueThreshold_) {
                  //printf("robotId %d   runNo %d   passNo %d  noveltyVal %f", robotId, runNo, passNo, noveltyValsByDistance[robotId][passNo][runNo][i]);
                  //printf("     FOUND \n ");
                  returnArray[passNo] += 1;
                  break;
               }
            }
         }
      }
   }
   return returnArray;
}

/**
 * Return 2D array of number of true positives in a certain pass and run, across all robots.
 * A true positive is when a robot that is near a novel object reports a novelty.
 *    0th dimension is passes
 *    1st dimension is runs
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 */
std::vector<std::vector<float>> DataProvider::GetNumOfNoveltyTruePositives(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, float noveltyDetectionRange_, Event::TYPE eventType_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_, MAX_NUM_OF_PASSES, noveltyValueThreshold_, noveltyDetectionRange_) != lastCalculationSetIdentifier_processNoveltyValues) { ProcessNoveltyValues(dataPath_, numOfRuns_, numOfRobots_, MAX_NUM_OF_PASSES, noveltyValueThreshold_, noveltyDetectionRange_); }
   return noveltyTruePositives;
}

/**
 * Return 2D array of number of false negatives in a certain pass and run, across all robots.
 * A false negative is when a robot that is near a novel object does not report a novelty.
 *    0th dimension is passes
 *    1st dimension is runs
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 */
std::vector<std::vector<float>> DataProvider::GetNumOfNoveltyFalseNegatives(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, float noveltyDetectionRange_, Event::TYPE eventType_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_, MAX_NUM_OF_PASSES, noveltyValueThreshold_, noveltyDetectionRange_) != lastCalculationSetIdentifier_processNoveltyValues) { ProcessNoveltyValues(dataPath_, numOfRuns_, numOfRobots_, MAX_NUM_OF_PASSES, noveltyValueThreshold_, noveltyDetectionRange_); }
   return noveltyFalseNegatives;
}

/**
 * Return 2D array of number of false positives in a certain pass and run, across all robots.
 * A false positive is when a robot that is not near a novel object reports a novelty.
 * A pass is this case is any time between last encountered the novel object (or novel object appearance) and next encounter.
 *    0th dimension is passes
 *    1st dimension is runs
 *
 * :param noveltyValueThreshold_: threshold above which a novelty value is considered as corresponding to novelty in the world
 */
std::vector<std::vector<float>> DataProvider::GetNumOfNoveltyFalsePositives(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, float noveltyDetectionRange_, Event::TYPE eventType_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_, MAX_NUM_OF_PASSES, noveltyValueThreshold_, noveltyDetectionRange_) != lastCalculationSetIdentifier_processNoveltyValues) { ProcessNoveltyValues(dataPath_, numOfRuns_, numOfRobots_, MAX_NUM_OF_PASSES, noveltyValueThreshold_, noveltyDetectionRange_); }
   return noveltyFalsePositives;
}



/* =====================================================================
                       TIME SERIES METRICS - NOVELTY
   ===================================================================== */

/**
 * Return 3D array of novelty data for a specific experiment, where:
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
std::vector<std::vector<std::vector<float>>> DataProvider::GetNoveltyValuesOverTime(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }
   return robotNoveltyValues;
}

/**
 * Return 3D array of winning neuron quantization error (distance from current input) data a specific experiment, where:
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
std::vector<std::vector<std::vector<float>>> DataProvider::GetWinningNeuronQuantizationErrorOverTime(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }
   return robotWinningNeuronQuantizationErrors;
}

/**
 * Return 3D array of winning neuron's neigbours quantization error (distance from current input) data for a specific experiment, where:
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
std::vector<std::vector<std::vector<float>>> DataProvider::GetWinningNeuronNeigboursQuantizationErrorOverTime(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }
   return robotWinningNeuronNeigbhoursQuantizationErrors;
}


/**
 * Return 3D array of start and end times where robots were around a novelty that has occurred, where:
 *    0th dimension is robot id
 *    1st dimension is run number
 *    2nd dimension is individual values, listed as startTime,endTime,startTime,endTime,...
 *
 * The maxDistance_ parameter is in metres.
 * The minDelayAfterEnvChange_ is the number of time ticks after an environment change when robot positions start being considered (e.g. in environments where robot teleports, this may need to be set to around 10)
 */
std::vector<std::vector<std::vector<float>>> DataProvider::GetRobotsAroundNoveltyTimes(std::string dataPath_, int numOfRuns_, int numOfRobots_, float maxDistance_, int minDelayAfterEnvChange_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }


   std::vector<std::vector<std::vector<float>>> returnArray(numOfRobots_, std::vector<std::vector<float>>(numOfRuns_, std::vector<float>(0,0)));

   int robotId, runNo, time;
   for (robotId = 0; robotId < numOfRobots_; robotId++) {
      for (runNo = 0; runNo < numOfRuns_; runNo++) {
         std::vector<Event>* envChangedEvents = GetEventsOfType(Event::TYPE::ENV_CHANGED, runNo);

         if (envChangedEvents == NULL) {
            printf("!!!!!! NO ENVIRONMENT CHANGED EVENTS for run no %d \n", runNo);
         }
         for (Event &event : *envChangedEvents) {
            //printf("robot id %d run no %d checking event at %f; %f time %d - %lu \n", robotId, runNo, event.position.GetX(), event.position.GetY(), event.time, robotPositions[robotId][runNo].size());
            int timeStartedFollowingRobot = -1;
            for (time = 0; time < (int)(robotPositions[robotId][runNo].size()); time++) {
               //-- only consider events that happened N time ticks after the event
               if (time >= event.time + minDelayAfterEnvChange_) { //
                  argos::CVector3* robotPosition = &robotPositions[robotId][runNo][time];
                  if ((*robotPosition - event.position).Length() <= maxDistance_) {
                     if (timeStartedFollowingRobot < 0) {
                        timeStartedFollowingRobot = time;
                     }
                  } else {
                     if (timeStartedFollowingRobot > 0) {
                        //printf("Robot %d around the event between %d and %d\n", robotId, timeStartedFollowingRobot, time);
                        returnArray[robotId][runNo].push_back(timeStartedFollowingRobot);
                        returnArray[robotId][runNo].push_back(time);
                        timeStartedFollowingRobot = -1;
                     }
                  }
               }
            }
         }
         //printf("DONE CHECKING EVENTS \n");
      }
   }

   return returnArray;
}


/**
 * Return 4D array of novelty values organised by distance from a novelty, where:
 *    0th dimension is robot id
 *    1st dimension is pass number
 *    2nd dimension is run number
 *    3rd dimension is individual values, listed by distance from novelty, in range [-noveltyDetectionRange_; noveltyDetectionRange_] (so middle values are for when robot at novelty)
 *
 * The maxDistance_ parameter is in metres.
 */
std::vector<std::vector<std::vector<std::vector<float>>>> DataProvider::GetNoveltyValuesByDistanceToNovelty(std::string dataPath_, int numOfRuns_, int numOfRobots_, int numOfPasses_, float maxDistance_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_, numOfPasses_, 999, maxDistance_) != lastCalculationSetIdentifier_processNoveltyValues) { ProcessNoveltyValues(dataPath_, numOfRuns_, numOfRobots_, numOfPasses_, 999, maxDistance_); }
   return noveltyValuesByDistanceToNovelty;
}

/* =====================================================================
                       TIME SERIES METRICS - EVENTS
   ===================================================================== */



/**
 * Return 2D array of times at which an event of a certain ID occurred where:
 *    0th dimension is run number
 *    1st dimension has the event times
 */
std::vector<std::vector<float>> DataProvider::GetEventTimes(std::string dataPath_, int numOfRuns_, int numOfRobots_, Event::TYPE eventType_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<std::vector<float>> returnArray;
   int runNo;
   for (runNo = 0; runNo < numOfRuns_; runNo++) {
      std::vector<float> eventTimes;
      std::vector<Event>* eventsVec = GetEventsOfType(eventType_, runNo);
      if (eventsVec != NULL) {
         for (Event &event : *eventsVec) {
             eventTimes.push_back(event.time);
         }
      }
      returnArray.push_back(eventTimes);
   }
   return returnArray;
}



/**
 * Return 3D array of times at which robots reached an instant waypoint where:
 *    0th dimension is robot id
 *    1st dimension is run number
 *    2nd dimension has the teleport times
 */
std::vector<std::vector<std::vector<float>>> DataProvider::GetRobotTeleportTimes(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<std::vector<std::vector<float>>> returnArray(numOfRobots_, std::vector<std::vector<float>>(numOfRuns_, std::vector<float>(0,0)));

   int robotId, runNo;
   for (robotId = 0; robotId < numOfRobots_; robotId++) {
      for (runNo = 0; runNo < numOfRuns_; runNo++) {
         std::vector<float> eventTimes;
         std::vector<Event>* eventsVec = GetEventsOfType(Event::TYPE::ROBOT_TELEPORTED, runNo);
         for (Event &event : *eventsVec) {
            returnArray[event.objectId][runNo].push_back(event.time);
         }
      }
   }
   return returnArray;
}

/**
 * Return 3D array of times at which robots reached a wapying of a certain id(string), where_
 *    0th dimension is robot id
 *    1st dimension is run number
 *    2nd dimension has the waypoint reached times
 */
std::vector<std::vector<std::vector<float>>> DataProvider::GetRobotReachedWaypointTimes(std::string dataPath_, int numOfRuns_, int numOfRobots_, std::string waypointId_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<std::vector<std::vector<float>>> returnArray(numOfRobots_, std::vector<std::vector<float>>(numOfRuns_, std::vector<float>(0,0)));

   int robotId, runNo;
   for (robotId = 0; robotId < numOfRobots_; robotId++) {
      std::vector<std::vector<float>> robotData;
      for (runNo = 0; runNo < numOfRuns_; runNo++) {
         std::vector<float> eventTimes;
         std::vector<Event>* eventsVec = GetEventsOfType(Event::TYPE::ROBOT_REACHED_WAYPOINT, runNo);
         for (Event &event : *eventsVec) {
            //printf("waypoint %s/n", event.waypointId);
            if (event.waypointId == waypointId_) {
               returnArray[event.objectId][runNo].push_back(event.time);
            }
         }
      }
   }
   return returnArray;
}

/* =====================================================================
                    HISTOGRAM METRICS - NEURAL NETWORK
   ===================================================================== */

/**
 * Return 2D array of bins for how many times a neuron was used where:
 *    0th dimension is robot id
 *    1st dimension is number of times used, with values representing percentage of neurons out of final network size
 *
 * Note: This function only works correctly for network that doesn't decrease in size!
 * Note: for each robot, the size of the array of counts may be different.
 */

std::vector<std::vector<float>> DataProvider::GetNeuronUsageAsBins(std::string dataPath_, int numOfRuns_, int numOfRobots_, int startTime_, int endTime_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<std::vector<float>> returnArray(numOfRobots_, std::vector<float>(1, 0)); // initially create space for counting 0 times fired.

   endTime_ = std::min(endTime_, (int)robotWinningNeurons[0][0].size());
   startTime_ = std::max(0, startTime_);

   int robotId, runNo, t, i, j, winningNeuron, numTimesFired, numOfNeurons;
   std::vector<int> neuronFireCounts;

   for (robotId=0; robotId<numOfRobots_; robotId++) {
      for (runNo=0; runNo < numOfRuns_; runNo++) {
         //printf("robotId %d runNo %d \n", robotId, runNo);
         neuronFireCounts.clear();
         for (t=startTime_; t<endTime_; t++) { // t=0 is time index, it really means t=1, i.e., the 1st time step
            //-- find out the total number of neurons at this time step. For time steps towards the end, the record does not exist, use the last known number of neurons
            if (t < (int)robotNumOfNeurons[robotId][runNo].size()) {
               numOfNeurons = robotNumOfNeurons[robotId][runNo][t];
            } else {
               numOfNeurons = robotNumOfNeurons[robotId][runNo][(int)robotNumOfNeurons[robotId][runNo].size()-1];
            }

            //-- make sure the neuronFireCounts is large enough
            if ((int)neuronFireCounts.size() < numOfNeurons) {
               //printf("neuron counts size %d   num of neurons %d \n",  (int)neuronFireCounts.size(), numOfNeurons );
               for (i=(int)neuronFireCounts.size(); i<numOfNeurons; i++) {
                  neuronFireCounts.push_back(0);
               }
               //std::cout << "t="<< t << " resizing neuron fire counts to  " << neuronFireCounts.size() << std::endl;
            }

            //-- increase fire count for the current winning neuron
            winningNeuron = robotWinningNeurons[robotId][runNo][t];
            if (winningNeuron >= 0) {
               neuronFireCounts[winningNeuron] ++;
               //printf(" t=%d increased fire count for neuron %d total count %d   %d - %d   %d \n", t, winningNeuron, neuronFireCounts[winningNeuron], startTime_, endTime_, (int)robotWinningNeurons[0][0].size());
            }
         }

         //-- organise neurons by fire count into bins, add up average over N runs and over maximum network size
         for (i=0; i<(int)neuronFireCounts.size(); i++) {
            numTimesFired = neuronFireCounts[i];
            if (numTimesFired >= (int)returnArray[robotId].size()) {
               for (j=(int)returnArray[robotId].size(); j<= numTimesFired; j++) {
                  returnArray[robotId].push_back(0);
               }
              //std::cout << "t="<< t << " resizing return array to  " << returnArray[robotId].size() << std::endl;
            }
            //returnArray[robotId][numTimesFired] += 100* ((1.0/numOfRuns_) / robotNumOfNeurons[robotId][runNo][robotNumOfNeurons[robotId][runNo].size()-1]);
            returnArray[robotId][numTimesFired] += 100* ((1.0/numOfRuns_) / robotNumOfNeurons[robotId][runNo][endTime_-1]);
            //printf(" neuron %d fired %d times. Total number of such neurons %f \n", i, numTimesFired, returnArray[robotId][neuronFireCounts[i]]);
         }
      }
   }


   return returnArray;
}


/* =====================================================================
                    TIME SERIES METRICS - NEURAL NETWORK
   ===================================================================== */

/**
 * Return 3D array of number of neurons for a specific experiment, where:
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
std::vector<std::vector<std::vector<float>>> DataProvider::GetNumOfNeuronsOverTime(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   int totalSimulationTime = robotNoveltyValues[0].size();
   std::vector<std::vector<std::vector<float>>> returnArray(numOfRobots_, std::vector<std::vector<float>>(totalSimulationTime, std::vector<float>(numOfRuns_, -1 )));

   //-- swap the 1st and 2nd dimension of the robotNumOfNeurons array
   for (int robotId=0; robotId<numOfRobots_; robotId++) {
      for (int t=0; t<totalSimulationTime; t++) {
         for (int runNo=0; runNo < numOfRuns_; runNo++) {
            //printf("t %d  arr size %lu   last val %f \n",t, robotNumOfNeurons[robotId][0].size(), robotNumOfNeurons[robotId][runNo][robotNumOfNeurons[robotId][runNo].size()-1]);
            if (t < (int)(robotNumOfNeurons[robotId][runNo].size())) {
               //-- values in robotNumOfNeurons still exist, because the last network grow event hasn't happened yet
               returnArray[robotId][t][runNo] = robotNumOfNeurons[robotId][runNo][t];
            } else {
               //-- last network growth event already happened, copy network size from last known
               returnArray[robotId][t][runNo] = robotNumOfNeurons[robotId][runNo][robotNumOfNeurons[robotId][runNo].size()-1];
            }
         }
      }
   }
   return returnArray;
}

/**
 * Return 3D array of how many neurons have been used (fired as winning) in each time step since time step N.
 * E.g. if there is total of 3 neurons and winning neurons for 5 time steps are 1,2,3,2,3, Then, for time step 4 (winning neuron 2), 3 neurons have been used since step 0, and 2 neurons since step 3.
 * When sinceTimeStep_ > 0, the values for t<sinceTimeStep_ are set to -1.
 *    0th dimension is robot id
 *    1st dimension is time
 *    2nd dimension is run number
 */
std::vector<std::vector<std::vector<float>>> DataProvider::GetNumOfUsedNeuronsOverTimeSince(std::string dataPath_, int numOfRuns_, int numOfRobots_, int sinceTime_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<std::vector<std::vector<float>>> returnArray(numOfRobots_, std::vector<std::vector<float>>(robotWinningNeurons[0][0].size(), std::vector<float>(numOfRuns_, -1 )));

   std::vector<float> neuronsUsedInThisRun;

   for (int robotId=0; robotId<numOfRobots_; robotId++) {

      for (int runNo=0; runNo < numOfRuns_; runNo++) {
         neuronsUsedInThisRun.clear();
         for (int t=sinceTime_; t<(int)robotWinningNeurons[robotId][0].size(); t++) {

            if (robotWinningNeurons[robotId][runNo][t] >= 0 && std::find(neuronsUsedInThisRun.begin(), neuronsUsedInThisRun.end(), robotWinningNeurons[robotId][runNo][t]) == neuronsUsedInThisRun.end()) {
               //-- neuron not used yet in this run
               neuronsUsedInThisRun.push_back(robotWinningNeurons[robotId][runNo][t]);
               //printf("robotId %d runNo %d t=%d new neuron used %f  total used %lu\n", robotId, runNo, t, robotWinningNeurons[robotId][runNo][t], neuronsUsedInThisRun.size());
            }

            //-- remember the number of neurons used for this time step
            returnArray[robotId][t][runNo] = neuronsUsedInThisRun.size();
         }
      }
   }

   return returnArray;
}


/* =====================================================================
                    SINGLE VALUE METRICS - NEURAL NETWORK
   ===================================================================== */

/**
 *  Return 1D array of maximum number of neurons, where
 *     0th dimension is values, includes those for different runs and different robots
 */
std::vector<float> DataProvider::GetNumOfNeurons(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<float> returnArray;

   for (int robotId = 0; robotId < numOfRobots_; robotId++) {
      for (int runNo = 0; runNo < numOfRuns_; runNo++) {
         std::vector<float>::iterator result;
         result = std::max_element(robotNumOfNeurons[robotId][runNo].begin(), robotNumOfNeurons[robotId][runNo].end());
         returnArray.push_back(*result);
      }
   }

   return returnArray;
}

/**
 *  Return 1D array of average total number of neuron in weights (averaged for each robot in each run), where
 *     0th dimension is values, includes those for different runs and different robots
 */
std::vector<float> DataProvider::GetNumOfInWeights(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<float> returnArray;

   for (int robotId = 0; robotId < numOfRobots_; robotId++) {
      for (int runNo = 0; runNo < numOfRuns_; runNo++) {
         returnArray.push_back(MathHelper::GetAverageOfAVector(robotNumOfInWeights[robotId][runNo], false));
      }
   }

   return returnArray;
}

/**
 *  Return 1D array of average number of neuron in weights per neuron (averaged for each robot in each run), where
 *     0th dimension is values, includes those for different runs and different robots
 */
std::vector<float> DataProvider::GetNumOfInWeightsPerNeuron(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<float> returnArray;
   std::vector<float> runArray;
   float inWeightsPerNeuron;
   for (int robotId = 0; robotId < numOfRobots_; robotId++) {
      for (int runNo = 0; runNo < numOfRuns_; runNo++) {
         runArray.clear();
         //-- get a vector of average num of in weights for each time step of this run and robot
         for (int t=0; t<(int)robotNumOfInWeights[robotId][runNo].size(); t++) {
            inWeightsPerNeuron = robotNumOfInWeights[robotId][runNo][t] / robotNumOfNeurons[robotId][runNo][t];
            //printf("robot %d runNo %d t=%d  num of neurons %f in weights %f adding %f\n", robotId, runNo, t, robotNumOfNeurons[robotId][runNo][t], robotNumOfInWeights[robotId][runNo][t], inWeightsPerNeuron);
            runArray.push_back(inWeightsPerNeuron);
         }
         //-- average the above
         returnArray.push_back(MathHelper::GetAverageOfAVector(runArray, false));
      }
   }

   return returnArray;
}

/**
 *  Return 1D array of average total number of inter-neuron connections (averaged for each robot in each run), where
 *     0th dimension is values, includes those for different runs and different robots
 */
std::vector<float> DataProvider::GetNumOfConnections(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<float> returnArray;

   for (int robotId = 0; robotId < numOfRobots_; robotId++) {
      for (int runNo = 0; runNo < numOfRuns_; runNo++) {
         returnArray.push_back(MathHelper::GetAverageOfAVector(robotNumOfConnections[robotId][runNo], false));
      }
   }

   return returnArray;
}

/**
 *  Return 1D array of average growth rate (% size increase) per second where
 *    0th dimension is values, includes those for different runs and different robots
 */
std::vector<float> DataProvider::GetNetworkGrowthRate(std::string dataPath_, int numOfRuns_, int numOfRobots_, int maxTime_) {
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   std::vector<float> returnArray;
   int maxTime;
   int timeStep = TIME_TICKS_PER_SECOND; // calculating growth rate per time step
   float growth;

   std::vector<float> growthRates;
   for (int runNo = 0; runNo < numOfRuns_; runNo++) {
      //-- for each robot, get when the network stopped growing
      std::vector<Event>* networkChangeEvents = GetEventsOfType(Event::TYPE::NN_CHANGED, runNo);
      std::vector<int> trainingTimePerRobot(numOfRobots_, 0);
      std::vector<int> lastNetworkSizePerRobot(numOfRobots_, -1);
      std::vector<int> areaUnderIncreasingCurve(numOfRobots_, 0);

      for (Event &event : *networkChangeEvents) {
         if (event.numOfNeurons != lastNetworkSizePerRobot[event.objectId]) {
            trainingTimePerRobot[event.objectId] = event.time;
            lastNetworkSizePerRobot[event.objectId] = event.numOfNeurons;
            areaUnderIncreasingCurve[event.objectId] += event.numOfNeurons;
         }
      }

      //-- get average growth rate per robot
      for (int robotId = 0; robotId < numOfRobots_; robotId++) {
         //printf("robot %d training time %d \n", robotId, trainingTimePerRobot[robotId]);
         //-- calculate growth rate as percentage of growth
         growthRates.clear();
         maxTime = std::min(maxTime_, std::min(trainingTimePerRobot[robotId] + timeStep, (int)(robotNumOfNeurons[robotId][runNo].size())));
         for (int t = timeStep; t < maxTime; t += timeStep) {
            growth = ((robotNumOfNeurons[robotId][runNo][t] - robotNumOfNeurons[robotId][runNo][t-timeStep])/robotNumOfNeurons[robotId][runNo][t-timeStep] ) *100;
            //printf("run %d robot %d t=%d   size before %f, now %f -> growth rate %f \n", runNo, robotId, t, robotNumOfNeurons[robotId][runNo][t-timeStep], robotNumOfNeurons[robotId][runNo][t],growth);
            growthRates.push_back(growth);
         }
         returnArray.push_back(MathHelper::GetAverageOfAVector(growthRates, true));

         //-- calculate area under the curve
         //returnArray.push_back(areaUnderIncreasingCurve[robotId]);

      }
   }

   return returnArray;
}


/* =====================================================================
                               SEARCH FUNCTIONS
   ===================================================================== */

/**
 * Get a pointer to events of a certain type that occured in a certain run
 */
std::vector<Event>* DataProvider::GetEventsOfType(Event::TYPE type_, int runNo_) {
   std::map<Event::TYPE,std::vector<Event>>::iterator it = eventsData[runNo_].find(type_);
   if (it != eventsData[runNo_].end()) {
      return &it->second;
   }
   //-- nothing found, return empty vector, using the map's entry for unknown type
   it = eventsData[runNo_].find(Event::TYPE::UNKNOWN);
   return &it->second;
}

/* =====================================================================
                               FILE AND HEAVY DATA PROCESSING
   ===================================================================== */


/**
 * Fill the noveltyValuesByDistanceToNovelty, falsePositives, falseNegatives and truePositives arrays. See array descriptions in the header file.
 * Note that a "pass" end when robot leaves the novel object (i.e. the object is behind robot more than detection range). So pass 0 = time between novel object is added and robot drives around it.
 * Pass 1 = between when robots first stops seeing object and until it leaves it again. So end of loop normally occurs in the middle of pass.
 * The maxDistance_ parameter is in metres.
 */
void DataProvider::ProcessNoveltyValues(std::string dataPath_, int numOfRuns_, int numOfRobots_, int numOfPasses_, float noveltyValueThreshold_, float maxDistance_) {

   printf("Processing novelty values...\n");
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_robotsLog) { ProcessRobotsLogData(dataPath_, numOfRuns_, numOfRobots_); }
   if (GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_) != lastCalculationSetIdentifier_events) { ProcessEventsData(dataPath_, numOfRuns_, numOfRobots_); }

   //-- remember what set of results was last calculated
   lastCalculationSetIdentifier_processNoveltyValues = GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_, numOfPasses_, noveltyValueThreshold_, maxDistance_);

   //-- initialise arrays
   int robotId, runNo, time, distanceMultiplier, binNo, passNo, nextRobotTeleportEventTime;
   int numOfBins = 181; // (numOfBins-1)/2 bins for before robot at novelty, (numOfBins-1)/2 bins for after robot at novelty, 1 bin for robot at novelty
   noveltyValuesByDistanceToNovelty.clear();
   for (robotId = 0; robotId < numOfRobots_; robotId++) {
      std::vector<std::vector<std::vector<float>>> vec(std::vector<std::vector<std::vector<float>>>(numOfPasses_, std::vector<std::vector<float>>(numOfRuns_, std::vector<float>(numOfBins,0))));
      noveltyValuesByDistanceToNovelty.push_back(vec);
   }

   noveltyTruePositives.clear();
   for (passNo = 0; passNo < numOfPasses_; passNo++) {
      std::vector<float> vec(std::vector<float>(numOfRuns_, 0));
      noveltyTruePositives.push_back(vec);
   }
   noveltyFalseNegatives.clear();
   for (passNo = 0; passNo < numOfPasses_; passNo++) {
      std::vector<float> vec(std::vector<float>(numOfRuns_, 0));
      noveltyFalseNegatives.push_back(vec);
   }
   noveltyFalsePositives.clear();
   for (passNo = 0; passNo < numOfPasses_; passNo++) {
      std::vector<float> vec(std::vector<float>(numOfRuns_, 0));
      noveltyFalsePositives.push_back(vec);
   }

   //--
   float distanceToNovelty, minDistance;
   bool truePositiveForThisPassReported;
   for (robotId = 0; robotId < numOfRobots_; robotId++) {
      for (runNo = 0; runNo < numOfRuns_; runNo++) {
         std::vector<Event>* envChangedEvents = GetEventsOfType(Event::TYPE::ENV_CHANGED, runNo);
         std::vector<Event>* robotsTeleportedEvents = GetEventsOfType(Event::TYPE::ROBOT_TELEPORTED, runNo);

         for (Event &event : *envChangedEvents) {
            //printf("checking event at %f; %f time %d \n", event.position.GetX(), event.position.GetY(), event.time);

            //-- if robots are supposed to teleport in this experiment, find the next teleport time after the environment changed.
            //    this will ensure that passes through environment are not missed in experiments like corridor, when an environment changes close to a robot that is about to teleport.
            nextRobotTeleportEventTime = -1;

            if (robotsTeleportedEvents->size() > 0) {
               for (Event &teleportEvent : *robotsTeleportedEvents) {
                  if (teleportEvent.time > event.time) {
                     nextRobotTeleportEventTime = teleportEvent.time;
                     break;
                  }
               }
               //printf("   next robot teleport event at %d\n", nextRobotTeleportEventTime);
            }


            distanceMultiplier = -1; // distances will be measured from -maxDistance_ to maxDistance_, so that a negative distance means before reaching the position of novelty
            minDistance = 9999;
            passNo = 0;
            std::vector<std::vector<std::vector<float>>> noveltyValsForThisEvent(numOfPasses_, std::vector<std::vector<float>>(numOfBins, std::vector<float>(0,0))); // will temporarily store bins of novelty values, each of which contains a vector, since multiple novelty values may belong to a single bin. Afterwards, averages will be taken for each bin and put into the return array.

            truePositiveForThisPassReported = false;
            for (time = 0; time < (int)(robotPositions[robotId][runNo].size()); time++) {
               //-- only consider events that happened after the event, and, if applicable, after robot teleported (see above)
               if (time >= event.time && (nextRobotTeleportEventTime < 0 || (nextRobotTeleportEventTime >= 0 && time >= nextRobotTeleportEventTime))) { //
                  argos::CVector3* robotPosition = &robotPositions[robotId][runNo][time];
                  distanceToNovelty = (*robotPosition - event.position).Length();
                  //printf(" d %f maxD %f    multip %d   current pass %d \n", distanceToNovelty, maxDistance_, distanceMultiplier, passNo);


                  if (distanceToNovelty <= maxDistance_) {
                     //-- novelty is within the robot detection range
                     if (distanceToNovelty < minDistance) {
                        //-- going downwards with distance, robot towards the novelty
                        distanceMultiplier = -1;
                        minDistance = distanceToNovelty;
                     } else {
                        //-- going upwards with distance, robot away from the novelty
                        distanceMultiplier = 1;
                     }
                     //-- apply the distance multiplier
                     distanceToNovelty = distanceMultiplier * distanceToNovelty;

                     binNo = round(((distanceToNovelty + maxDistance_) / (maxDistance_*2)) * (numOfBins-1));
                     if (robotNoveltyValues[robotId][time][runNo] > 0) {
                        //printf("pass no %d   distance to novelty: %f binNo: %d    time %d, robot novelty val %f \n", passNo, distanceToNovelty, binNo, time, robotNoveltyValues[robotId][time][runNo]);
                     }
                     if (binNo >= 0 && binNo < numOfBins) {
                        noveltyValsForThisEvent[passNo][binNo].push_back(robotNoveltyValues[robotId][time][runNo]);
                     } else {
                        printf("  !!!! could not assign distance to novelty: %f into a bin (bin no %d, pass no %d, event at pos %f;%f \n )", distanceToNovelty, binNo, passNo, event.position.GetX(), event.position.GetY());
                     }

                     //-- is this a true positive?
                     if (robotNoveltyValues[robotId][time][runNo] >= noveltyValueThreshold_ && !truePositiveForThisPassReported) {
                        noveltyTruePositives[passNo][runNo] += 1;
                        truePositiveForThisPassReported = true;
                     }


                  } else {

                     //-- robot should outside of novel object detection range
                     if (robotNoveltyValues[robotId][time][runNo] > noveltyValueThreshold_) {
                        noveltyFalsePositives[passNo][runNo] += 1;
                        //printf("robotId %d runNo %d passNo %d novelty false positive (%f) detected at t=%d, distance from novel object=%f \n",robotId, runNo, passNo, robotNoveltyValues[robotId][time][runNo],time,distanceToNovelty);
                     }

                     if (distanceMultiplier == 1) {
                        //-- was there a false negative? (i.e. a positive not reported)?
                        if (!truePositiveForThisPassReported) {
                           noveltyFalseNegatives[passNo][runNo] += 1;
                        }

                        //-- reset for next pass
                        minDistance = 9999;
                        passNo += 1;
                        distanceMultiplier = -1;
                        truePositiveForThisPassReported = false;

                        if (passNo >= numOfPasses_) {
                           break;
                        }
                     }
                  }

               }
            }

            //-- by this point, all novelty data has been taken but some bins contain multiple novelty values. Average them into the return array
            for (passNo = 0; passNo < numOfPasses_; passNo++) {
               for (binNo = 0; binNo < numOfBins; binNo++) {
                  noveltyValuesByDistanceToNovelty[robotId][passNo][runNo][binNo] = MathHelper::GetMaxOfAVector(noveltyValsForThisEvent[passNo][binNo]);
                  if (passNo == 0) {
                     //printf("  pass %d bin %d     final val %f\n ", passNo, binNo, noveltyValuesByDistanceToNovelty[robotId][passNo][runNo][binNo]);
                  }
               }
            }
         }
      }
   }
}

/**
 * Go through robots log for all runs in a given dataPath_ and store various data in various arrays.
 * Fills the robotNoveltyValues, robotPositions, robotWinningNeurons arrays.
 */
void DataProvider::ProcessRobotsLogData(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   printf("Processing robots log data...\n");

   //-- remember what set of results was last calculated
   lastCalculationSetIdentifier_robotsLog = GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_);

   //-- initialise various arrays
   int robotId, runNo, lineNo, time;
   robotNoveltyValues.clear();
   robotPositions.clear();
   robotWinningNeurons.clear();
   robotWinningNeuronQuantizationErrors.clear();
   robotWinningNeuronNeigbhoursQuantizationErrors.clear();
   for (robotId = 0; robotId < numOfRobots_; robotId++) {
      std::vector<std::vector<float>> arr;
      robotNoveltyValues.push_back(arr);

      std::vector<std::vector<argos::CVector3>> arr2(numOfRuns_, std::vector<argos::CVector3>(0,argos::CVector3()));
      robotPositions.push_back(arr2);

      std::vector<std::vector<float>> arr3(numOfRuns_, std::vector<float>(0,-1));
      robotWinningNeurons.push_back(arr3);

      std::vector<std::vector<float>> arr4;
      robotWinningNeuronQuantizationErrors.push_back(arr4);

      std::vector<std::vector<float>> arr5;
      robotWinningNeuronNeigbhoursQuantizationErrors.push_back(arr5);
   }



   //-- go through individual run files
   std::vector<std::vector<float>> fileData;
   for (runNo = 0; runNo < numOfRuns_; runNo++) {
      std::stringstream filePathStream;
      filePathStream << dataPath_ << "/run" << std::to_string(runNo) << "_robotsLog.txt";
      fileData.clear();
      IOHelper::FileToVector(filePathStream.str(), fileData, true);
      //printf("Processing ... \n");
      for (lineNo = 0; lineNo < (int)(fileData.size()); lineNo++) {
         if (runNo == 0) {
            //printf("Creating structure \n");
            //-- create the structure in the arrays that have 1st dimension time for this time step (line), for all runs
            for (robotId = 0; robotId < numOfRobots_; robotId ++) {
               std::vector<float> timeStepArr(numOfRuns_, -1);
               robotNoveltyValues[robotId].push_back(timeStepArr);

               std::vector<float> timeStepArr2(numOfRuns_, -1);
               robotWinningNeuronQuantizationErrors[robotId].push_back(timeStepArr2);

               std::vector<float> timeStepArr3(numOfRuns_, -1);
               robotWinningNeuronNeigbhoursQuantizationErrors[robotId].push_back(timeStepArr3);
            }
         }
         //printf(" line %d array length %lu \n",lineNo, robotNoveltyValues[0].size());
         robotId = fileData[lineNo][ROBOTSLOG_ROBOT_ID];
         time = fileData[lineNo][ROBOTSLOG_TIME];
         //printf("t=%f  robot id %d \n", time, robotId );
         if (robotId < numOfRobots_ && time >= 0) {
            robotNoveltyValues[robotId][time][runNo] = fileData[lineNo][ROBOTSLOG_NOVELTY_VAL];
            //-- adjust very small values to 0, because of python plotting
            if (robotNoveltyValues[robotId][time][runNo] < 0.05) {
               robotNoveltyValues[robotId][time][runNo] = 0;
            }

            robotWinningNeuronQuantizationErrors[robotId][time][runNo] = fileData[lineNo][ROBOTSLOG_WINNING_NEURON_QUANTIZATION_ERROR];
            robotWinningNeuronNeigbhoursQuantizationErrors[robotId][time][runNo] = fileData[lineNo][ROBOTSLOG_WINNING_NEURON_NEIGHBOURS_QUANTIZATION_ERROR];
         }

         //-- save the robot's data
         robotPositions[robotId][runNo].push_back(argos::CVector3(fileData[lineNo][ROBOTSLOG_POSITION_X],fileData[lineNo][ROBOTSLOG_POSITION_Y],0));
         robotWinningNeurons[robotId][runNo].push_back(fileData[lineNo][ROBOTSLOG_WINNING_NEURON]);


      }
   }
}

/**
 * Go through robots log for all runs in a given dataPath_ and store various data in various arrays.
 * Fills the eventsData, robotNumOfNeurons arrays.
 */
void DataProvider::ProcessEventsData(std::string dataPath_, int numOfRuns_, int numOfRobots_) {

   printf("Processing events data...\n");

   //-- remember what set of results was last calculated
   lastCalculationSetIdentifier_events = GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, numOfRobots_);

   //-- initialise various arrays
   int runNo, lineNo, robotId, t, objectId;


   Event::TYPE eventType;
   eventsData.clear();
   for (runNo = 0; runNo < numOfRuns_; runNo++) {
      std::map<Event::TYPE, std::vector<Event>> map;
      eventsData.push_back(map);
   }

   robotNumOfNeurons.clear();
   robotNumOfConnections.clear();
   robotNumOfInWeights.clear();
   for (robotId = 0; robotId < numOfRobots_; robotId++) {
      std::vector<std::vector<float>> arr(numOfRuns_, std::vector<float>(0,-1));
      robotNumOfNeurons.push_back(arr);
      std::vector<std::vector<float>> arr2(numOfRuns_, std::vector<float>(0,-1));
      robotNumOfConnections.push_back(arr2);
      std::vector<std::vector<float>> arr3(numOfRuns_, std::vector<float>(0,-1));
      robotNumOfInWeights.push_back(arr3);
   }


   std::pair<std::map<Event::TYPE,std::vector<Event>>::iterator,bool> mapIterator;
   std::map<Event::TYPE,std::vector<Event>>::iterator it;

   //-- go through individual run files, search for the relevant events
   for (runNo = 0; runNo < numOfRuns_; runNo++) {
      std::vector<int> lastNNNumOfNeurons(numOfRobots_, 0);
      std::vector<int> lastNNNumOfInWeights(numOfRobots_, 0);
      std::vector<int> lastNNNumOfConnections(numOfRobots_, 0);
      std::vector<int> lastNNChangedEventTime(numOfRobots_, 0);

      //-- insert empty vector that maps onto event unknown
      mapIterator = eventsData[runNo].insert(std::pair<Event::TYPE,std::vector<Event>>(Event::TYPE::UNKNOWN,std::vector<Event>(0,Event())));

      //-- insert vectors with events based on the log file
      std::stringstream filePathStream;
      filePathStream << dataPath_ << "/run" << std::to_string(runNo) << "_events.txt";
      std::vector<std::vector<std::string>> fileData;
      IOHelper::FileToVector(filePathStream.str(), fileData, true);

      for (lineNo = 0; lineNo < (int)(fileData.size()); lineNo++) {
         //printf("LINE %d / %lu \n",lineNo, fileData.size());
         /*for (int i=0; i<fileData[lineNo].size(); i++) {
            printf("%f ", fileData[lineNo][i]);
         }*/
         eventType = Event::TYPE((std::stoi(fileData[lineNo][EVENTSLOG_EVENT_TYPE])));
         objectId = std::stoi(fileData[lineNo][EVENTSLOG_OBJECT_ID]);
         //printf("run %d  line %d  eventType %d", runNo, lineNo, eventType);

         //-- get the vector of events of this type from the events map or create it
         mapIterator = eventsData[runNo].insert(std::pair<Event::TYPE,std::vector<Event>>(eventType,std::vector<Event>(0,Event())));
         it = mapIterator.first;
         std::vector<Event>* vec = &it->second;
         //-- add new event to the vector
         if (eventType == Event::TYPE::NN_CHANGED) {
            vec->push_back(Event(eventType, std::stoi((fileData[lineNo][EVENTSLOG_TIME])), objectId, std::stoi((fileData[lineNo][EVENTSLOG_NN_NUM_OF_NEURONS])), std::stoi((fileData[lineNo][EVENTSLOG_NN_NUM_OF_IN_WEIGHTS])), std::stoi((fileData[lineNo][EVENTSLOG_NN_NUM_OF_CONNECTIONS]))));
         } else if (eventType == Event::TYPE::ROBOT_REACHED_WAYPOINT) {
            vec->push_back(Event(eventType, std::stoi((fileData[lineNo][EVENTSLOG_TIME])), objectId, argos::CVector3(std::stof(fileData[lineNo][EVENTSLOG_POSITION_X]), std::stof(fileData[lineNo][EVENTSLOG_POSITION_Y]),0), fileData[lineNo][EVENTSLOG_WAYPOINT_ID] ));
         } else {
            vec->push_back(Event(eventType, std::stoi((fileData[lineNo][EVENTSLOG_TIME])), objectId, argos::CVector3(std::stof(fileData[lineNo][EVENTSLOG_POSITION_X]), std::stof(fileData[lineNo][EVENTSLOG_POSITION_Y]),0)));
         }

         //-- process events related to network changes
         if (eventType == Event::TYPE::NN_CHANGED) {
            //-- fill in previous num of neurons from last time until eventTime-1
            for (t=lastNNChangedEventTime[objectId]+1; t<std::stoi(fileData[lineNo][EVENTSLOG_TIME]); t++) {
               robotNumOfNeurons[objectId][runNo].push_back(lastNNNumOfNeurons[objectId]);
               robotNumOfInWeights[objectId][runNo].push_back(lastNNNumOfInWeights[objectId]);
               robotNumOfConnections[objectId][runNo].push_back(lastNNNumOfConnections[objectId]);
            }
            //printf("Filling network size  %d numInWeights %d numConnections %d from t=%d to %d (robot %d  final array size %lu) \n",lastNNNumOfNeurons[objectId], lastNNNumOfInWeights[objectId], lastNNNumOfConnections[objectId], lastNNChangedEventTime[objectId]+1, std::stoi(fileData[lineNo][EVENTSLOG_TIME])-1,objectId, robotNumOfNeurons[objectId][runNo].size());

            //-- fill in current num of neurons for the eventTime. This step assures that the last size is recorded at least once.
            lastNNChangedEventTime[objectId] = std::stoi(fileData[lineNo][EVENTSLOG_TIME]);
            lastNNNumOfNeurons[objectId] = std::stoi(fileData[lineNo][EVENTSLOG_NN_NUM_OF_NEURONS]);
            lastNNNumOfInWeights[objectId] = std::stoi(fileData[lineNo][EVENTSLOG_NN_NUM_OF_IN_WEIGHTS]);
            lastNNNumOfConnections[objectId] = std::stoi(fileData[lineNo][EVENTSLOG_NN_NUM_OF_CONNECTIONS]);
            robotNumOfNeurons[objectId][runNo].push_back(lastNNNumOfNeurons[objectId]);
            robotNumOfInWeights[objectId][runNo].push_back(lastNNNumOfInWeights[objectId]);
            robotNumOfConnections[objectId][runNo].push_back(lastNNNumOfConnections[objectId]);
            //printf("Filling network size  %d numInWeights %d numConnections %d from t=%d to %d (robot %d  final array size %lu) \n",lastNNNumOfNeurons[objectId], lastNNNumOfInWeights[objectId], lastNNNumOfConnections[objectId], lastNNChangedEventTime[objectId], std::stoi(fileData[lineNo][EVENTSLOG_TIME]),objectId, robotNumOfNeurons[objectId][runNo].size());
         }
      }
   }
}


/**
 * Generate a string combining parameters like dataPath, numOfRuns, etc.
 */
std::string DataProvider::GenerateCalculationSetIdentifier(std::string dataPath_, int numOfRuns_) {
   return GenerateCalculationSetIdentifier(dataPath_, numOfRuns_, 0);
}

std::string DataProvider::GenerateCalculationSetIdentifier(std::string dataPath_, int numOfRuns_, int numOfRobots_) {
   std::stringstream stream;
   stream << dataPath_ << "-" << std::to_string(numOfRuns_) << "runs-" << numOfRobots_ << "robots";
   return stream.str();
}

std::string DataProvider::GenerateCalculationSetIdentifier(std::string dataPath_, int numOfRuns_, int numOfRobots_, float valX_, float valY_, float valZ_) {
   std::stringstream stream;
   stream << dataPath_ << "-" << std::to_string(numOfRuns_) << "runs-" << numOfRobots_ << "robots-" << valX_ << "-" << valY_ << "-" << valZ_;
   return stream.str();
}

