/* ================================================================
 * ================================================================
 *
 * The main data provider class, called from gaData_module.
 *
 * Author: L. Pitonakova (http://lenkaspace.net)
 * License: GNU General Public License. Please credit me when using my work.
 *
 * ================================================================
 * ================================================================
 */

#ifndef DATAPROVIDER_H_
#define DATAPROVIDER_H_

#include <string>
#include <vector>
#include <event.h>
#include <map>

#include <argos3/core/utility/math/vector3.h>

//------ define what different indexes mean in data files
#define ROBOTSLOG_TIME 0
#define ROBOTSLOG_ROBOT_ID 1
#define ROBOTSLOG_POSITION_X 2
#define ROBOTSLOG_POSITION_Y 3
#define ROBOTSLOG_STATE 4
#define ROBOTSLOG_NOVELTY_VAL 5
#define ROBOTSLOG_WINNING_NEURON 6
#define ROBOTSLOG_WINNING_NEURON_QUANTIZATION_ERROR 7
#define ROBOTSLOG_WINNING_NEURON_NEIGHBOURS_QUANTIZATION_ERROR 8

#define EVENTSLOG_TIME 0
#define EVENTSLOG_EVENT_TYPE 1
#define EVENTSLOG_OBJECT_ID 2
#define EVENTSLOG_POSITION_X 3   // for events of type ENV_CHANGED, ROBOT_REACHED_WAYPOINT and ROBOT_TELEPORTED
#define EVENTSLOG_POSITION_Y 4   // for events of type ENV_CHANGED, ROBOT_REACHED_WAYPOINT and ROBOT_TELEPORTED
#define EVENTSLOG_WAYPOINT_ID 5 // for events of type ROBOT_REACHED_WAYPOINT
#define EVENTSLOG_NN_NUM_OF_NEURONS 3 // for events of type NN_CHANGED
#define EVENTSLOG_NN_NUM_OF_IN_WEIGHTS 4 // for events of type NN_CHANGED
#define EVENTSLOG_NN_NUM_OF_CONNECTIONS 5 // for events of type NN_CHANGED


#define MIN_DELAY_AFTER_ENV_CHANGE 10  // used in GetRobotsAroundNoveltyTimes

#define TIME_TICKS_PER_SECOND 10
#define MAX_NUM_OF_PASSES 40

class DataProvider {
public:
   DataProvider();
   virtual ~DataProvider();

   std::vector<float> GetNetworkTrainingTime(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, int startTime_, int endTime_);
   std::vector<float> GetNumOfRunsNovelObjectReported(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, float noveltyDetectionRange_, Event::TYPE eventType_);

   std::vector<std::vector<float>> GetNumOfNoveltyTruePositives(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, float noveltyDetectionRange_, Event::TYPE eventType_);
   std::vector<std::vector<float>> GetNumOfNoveltyFalseNegatives(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, float noveltyDetectionRange_, Event::TYPE eventType_);
   std::vector<std::vector<float>> GetNumOfNoveltyFalsePositives(std::string dataPath_, int numOfRuns_, int numOfRobots_, float noveltyValueThreshold_, float noveltyDetectionRange_, Event::TYPE eventType_);



   std::vector<std::vector<std::vector<float>>> GetNoveltyValuesOverTime(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<std::vector<std::vector<float>>> GetWinningNeuronQuantizationErrorOverTime(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<std::vector<std::vector<float>>> GetWinningNeuronNeigboursQuantizationErrorOverTime(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<std::vector<float>> GetEventTimes(std::string dataPath_, int numOfRuns_, int numOfRobots_, Event::TYPE eventType_);
   std::vector<std::vector<std::vector<float>>> GetRobotTeleportTimes(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<std::vector<std::vector<float>>> GetRobotReachedWaypointTimes(std::string dataPath_, int numOfRuns_, int numOfRobots_, std::string waypointId_);
   std::vector<std::vector<std::vector<float>>> GetRobotsAroundNoveltyTimes(std::string dataPath_, int numOfRuns_, int numOfRobots_, float maxDistance_, int minDelayAfterEnvChange_);
   std::vector<std::vector<std::vector<std::vector<float>>>> GetNoveltyValuesByDistanceToNovelty(std::string dataPath_, int numOfRuns_, int numOfRobots_, int numOfPasses_, float maxDistance_);

   std::vector<std::vector<float>> GetNeuronUsageAsBins(std::string dataPath_, int numOfRuns_, int numOfRobots_, int startTime_, int endTime_);

   std::vector<std::vector<std::vector<float>>> GetNumOfNeuronsOverTime(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<std::vector<std::vector<float>>> GetNumOfUsedNeuronsOverTimeSince(std::string dataPath_, int numOfRuns_, int numOfRobots_, int sinceTime_);

   std::vector<float> GetNumOfNeurons(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<float> GetNumOfInWeights(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<float> GetNumOfInWeightsPerNeuron(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<float> GetNumOfConnections(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::vector<float> GetNetworkGrowthRate(std::string dataPath_, int numOfRuns_, int numOfRobots_, int maxTime_);


private:
   void ProcessNoveltyValues(std::string dataPath_, int numOfRuns_, int numOfRobots_, int numOfPasses_, float noveltyValueThreshold_, float maxDistance_);
   void ProcessRobotsLogData(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   void ProcessEventsData(std::string dataPath_, int numOfRuns_, int numOfRobots_);

   std::string GenerateCalculationSetIdentifier(std::string dataPath_, int numOfRuns_);
   std::string GenerateCalculationSetIdentifier(std::string dataPath_, int numOfRuns_, int numOfRobots_);
   std::string GenerateCalculationSetIdentifier(std::string dataPath_, int numOfRuns_, int numOfRobots_, float valX_, float valY_, float valZ_);

   std::vector<Event>* GetEventsOfType(Event::TYPE type_, int runNo_);


   std::vector<std::vector<std::vector<std::vector<float>>>> noveltyValuesByDistanceToNovelty;
   std::vector<std::vector<float>> noveltyTruePositives;
   std::vector<std::vector<float>> noveltyFalseNegatives;
   std::vector<std::vector<float>> noveltyFalsePositives;

   std::vector<std::vector<std::vector<float>>> robotNoveltyValues;  // 0th dimension is robot id, 1st dimension is time, 2nd dimension is run number

   std::vector<std::vector<std::vector<float>>> robotWinningNeuronQuantizationErrors; // 0th dimension is robot id, 1st dimension is time, 2nd dimension is run number
   std::vector<std::vector<std::vector<float>>> robotWinningNeuronNeigbhoursQuantizationErrors; // 0th dimension is robot id, 1st dimension is time, 2nd dimension is run number

   std::vector<std::vector<std::vector<argos::CVector3>>> robotPositions; // 0th dimension is robot id, 1st dimension is run number, 2nd dimension is time
   std::vector<std::vector<std::vector<float>>> robotWinningNeurons; // 0th dimension is robot id, 1st dimension is run number, 2nd dimension is time
   std::vector<std::vector<std::vector<float>>> robotNumOfNeurons; // 0th dimension is robot id, 1st dimension is run number, 2nd dimension is time
   std::vector<std::vector<std::vector<float>>> robotNumOfInWeights; // 0th dimension is robot id, 1st dimension is run number, 2nd dimension is time
   std::vector<std::vector<std::vector<float>>> robotNumOfConnections; // 0th dimension is robot id, 1st dimension is run number, 2nd dimension is time
   std::vector<std::map<Event::TYPE, std::vector<Event>>> eventsData; // 0th dimension is run number, 1st dimension is map indexed by event type, 2nd dimension is individual events

   //-- every time something is recalculated, these is set based on path and parameters. Then data can be reused.
   std::string lastCalculationSetIdentifier_processNoveltyValues;
   std::string lastCalculationSetIdentifier_robotsLog;
   std::string lastCalculationSetIdentifier_events;

};

#endif /* DATAPROVIDER_H_ */
