####################################################################
##
## Code for creating all figures and stats presented in paper
## Pitonakova and Bullock, ﻿Robustness and Structural Plasticity of Grow-When-Required Neural Networks, submitted to ICRA 2019.
##
## Author: L. Pitonakova (http://lenkaspace.net)
## License: GNU General Public License. Please credit me when using my work.
##
###################################################################

import gaData_module;
from pyCreeper import crGraphs;
from pyCreeper import crGraphStyle;
from pyCreeper import crData;
import math;
import helpers;
import constants;
import visualization;
import paths;


#-- specify which plots to make based on paper sections
DO_NETWORK_MODES_ALL_NETWORKS = True;
DO_NETWORK_MODES_SELECTED_NETWORKS = True;
DO_NETWORK_ROBUSTNESS_PARAMETERS = True;
DO_NETWORK_ROBUSTNESS_ROBOT_SPEED = True;
DO_NETWORK_ROBUSTNESS_NOISE = True;
#----

def resetPlotConfig():
   config_ = visualization.PlotConfig();
   config_.style.lineWidth = 3;
   config_.numOfPassesToShowInDetail = 1;
   config_.numOfRuns = 50;
   return config_;

config = resetPlotConfig()

MAIN_OUTPUT_DIRECTORY = paths.BASE_FILE_PATH_OUTPUT_IMAGES + "/__images_cp01";
helpers.createFolder(MAIN_OUTPUT_DIRECTORY);



PARAMETER_SET_LABELS = ["Specialist", "Generalist", "Plastic"]


def getSelectedNetworkDirectories(environmentName_):

   if ("corridor" in environmentName_):
      specialist = "inw12_at0.8";
      generalist = "inw39_at0.2";
   else:
      specialist = "inw9_at0.7";
      generalist = "inw30_at0.2";


   return ["s1_x2_" + environmentName_ + "/" + specialist + "_rms20",
           "s1_x2_" + environmentName_ + "/" + generalist + "_rms20",
           "s1_x2_" + environmentName_ + "/inwAdaptive_at0.2_rms20"];


#------------------------------------
# NETWORK MODES
#------------------------------------

if (DO_NETWORK_MODES_ALL_NETWORKS):
   config = resetPlotConfig()
   for analysisName in ["corridor", "room"]:
      OUTPUT_DIRECTORY = MAIN_OUTPUT_DIRECTORY + "/networkModes";

      ACTIVATION_THRESHOLD_VALS = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]; #
      #ACTIVATION_THRESHOLD_VALS = [0.1, 0.2, 0.8, 0.9]; #
      IN_WEIGHT_VALS = [9, 12, 18, 24, 30, 39, 60, "Adaptive"];
      IN_WEIGHT_VAL_LABELS = []
      for inwVal in IN_WEIGHT_VALS:
         if (inwVal == "Adaptive"):
            IN_WEIGHT_VAL_LABELS.append("Plastic");
         else:
            IN_WEIGHT_VAL_LABELS.append(inwVal);

      averagePassLengths = [[] for p in range(len(IN_WEIGHT_VALS))];
      noveltyOccurrenceMinTimes = [[] for p in range(len(IN_WEIGHT_VALS))];
      initialNetworkTrainingTimeData = [[] for inw in range(len(IN_WEIGHT_VALS))];

      numOfNoveltyTruePositivesData = [[] for inw in range(len(IN_WEIGHT_VALS))];      # each data point is collated number of true positives in a certain pass and run, across all robots. A true positive is when a robot that is near a novel object reports a novelty
      numOfNoveltyFalsePositivesData = [[] for inw in range(len(IN_WEIGHT_VALS))];    # each data point is collated number of false positives in a certain pass and run, across all robots. A false positive is when a robot that is not near a novel object reports a novelty

      for inw in range(len(IN_WEIGHT_VALS)):
         inwVal = IN_WEIGHT_VALS[inw];
         for at in range(len(ACTIVATION_THRESHOLD_VALS)):
            atVal = ACTIVATION_THRESHOLD_VALS[at];
            config.dataDirectoryPath = paths.BASE_FILE_PATH_DATA + "/s1_x2_" + analysisName + "/inw" + str(inwVal) + "_at" + str(atVal) + "_rms20";

            envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
            averagePassLength, maxPassLength = helpers.calculatePassLength(config);
            averagePassLengths[inw].append(averagePassLength);

            noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);
            noveltyOccurrenceMinTimes[inw].append(noveltyOccurrenceMinTime);


            #---- get single measure data
            initialNetworkTrainingTimeData[inw].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));
            numOfNoveltyTruePositivesData[inw].append(gaData_module.getNumOfNoveltyTruePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));
            numOfNoveltyFalsePositivesData[inw].append(gaData_module.getNumOfNoveltyFalsePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));


      #------ plot single measure data
      config.style.markers =     ['ks',      'ks',       'ks',       'ks',    'ks',       'ks',       'ks',       'ks'];
      config.style.colors =      ['#FF8000', '#00CD00',  'green',    'blue',  '#B23AEE',  '#8B4C39',  '#282828',  '#EE0000' ];
      config.style.lineStyles =  ['-',       '-',        '-',        '-',     '-',        '-',        '-',        '--'];
      config.style.numOfLegendColumns = 2;
      config.outputDirectoryPath = OUTPUT_DIRECTORY + "/" + analysisName + "_"
      xAxisLabel = "Activation threshold ($\Theta_A$)";
      xAxisTicks = ACTIVATION_THRESHOLD_VALS;
      legendLabels = IN_WEIGHT_VAL_LABELS;
      legendTitle = "$||w_c||$"

      if ("corridor" in analysisName):
         config.style.legendPosition = crGraphStyle.LEGEND_POSITION.NONE;
      else:
         config.style.legendPosition = crGraphStyle.LEGEND_POSITION.BEST;

      #-- network training time
      visualization.plotNetworkTrainingTimeByPass(config, initialNetworkTrainingTimeData, averagePassLengths, [[0 for v in range(len(ACTIVATION_THRESHOLD_VALS))] for p in range(len(IN_WEIGHT_VALS))], legendLabels, xAxisLabel, xAxisTicks, legendTitle_=legendTitle, showBoxPlots_=False);

      #-- performance
      visualization.plotNoveltyReporting(config, numOfNoveltyTruePositivesData, "true positives", legendLabels, xAxisLabel, xAxisTicks, legendTitle_=legendTitle, convertValuesToPercentages_=True);

      #yMax = crGraphs.INVALID_VALUE;
      #brokenPlotYMax = crGraphs.INVALID_VALUE;
      #brokenPlotYCoord = crGraphs.INVALID_VALUE;

      #if ("corridor" in analysisName):
      #   brokenPlotYMax = 9.0;
      #   brokenPlotYCoord = 2.0;
      #if ("room" in analysisName):
      #   yMax = 2.5;
      #visualization.plotNoveltyReporting(config, numOfNoveltyFalsePositivesData, "false positives", [], xAxisLabel, xAxisTicks, yMax_ = yMax, brokenLinePlot_yMax_=brokenPlotYMax,brokenLinePlot_yCoord=brokenPlotYCoord);


#------------------------------------
# NETWORK MODES - SELECTED NETWORKS PLOTS
#------------------------------------
if (DO_NETWORK_MODES_SELECTED_NETWORKS):
   config = resetPlotConfig()
   OUTPUT_DIRECTORY = MAIN_OUTPUT_DIRECTORY + "/networkModes";
   for analysisName in ["corridor", "room"]: #
      networkDirectories = getSelectedNetworkDirectories(analysisName);

      robotsWinningNeuronErrorValues = [];
      robotsWinningNeuronNeigboursErrorValues = [];
      envChangedTimes = [];
      networkTrainingTimes = [];

      for dir in networkDirectories:
         config.dataDirectoryPath = paths.BASE_FILE_PATH_DATA + "/" + dir;

         envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
         envChangedTimes.append(envChangedTimesData)
         averagePassLength, maxPassLength = helpers.calculatePassLength(config);
         noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);

         networkTrainingTimeAverage = (int)(math.ceil(crData.getAverageOfAList(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime))));

         #------ calculate and write to file the network size after training
         networkTrainingTimesData = gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime);
         networkTrainingTimes.append(networkTrainingTimesData);
         robotsNetworkSize = gaData_module.getNumOfNeuronsOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots);

         networkSizesAfterTraining = [];
         for runNo in range(config.numOfRuns):
            networkSizesAfterTraining.append(robotsNetworkSize[0][(int)(networkTrainingTimesData[runNo]) + 10][runNo]);
         outputStr = dir + ": Size after training average: " + str(crData.getAverageOfAList(networkSizesAfterTraining)) + "  median: " + str(crData.getMedianOfAList(networkSizesAfterTraining)) + "\n";

         fileHandle = 'a';
         if (dir == networkDirectories[0]):
            fileHandle = 'w';
         f = open(OUTPUT_DIRECTORY + "/" + analysisName + "_networkSizeStats.txt",fileHandle)
         f.write(outputStr);
         f.close()

         #------ do the neuron fire counts histograms, for each parameter set separately
         config.outputDirectoryPath = OUTPUT_DIRECTORY + "/neuralNetworkHistograms/" + dir;

         if ("corridor" in analysisName):
            yMax = 60;
            if ("at0.2" in config.dataDirectoryPath):
               yMax = 30;
         else:
            yMax = crGraphs.INVALID_VALUE;
         neuronUsageInBins = gaData_module.getNeuronUsageAsBins(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,networkTrainingTimeAverage,(int)(networkTrainingTimeAverage + averagePassLength));

         config.style.figureSize = (1.7,3.7)
         config.style.xOffset = 0.12;
         fileName = "neuronUsage_beforeNovelty_trained_firstPass";
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, firstNBins_=10, yMax_=yMax);

         config.style.figureSize = (3.5,3.7)
         config.style.xOffset = 0.0;
         visualization.plotNNUsageHistogram(config, neuronUsageInBins, fileName, yMax_=yMax);


         #------ neuron quantization error
         robotsWinningNeuronErrorValues.append(gaData_module.getWinningNeuronQuantizationErrorOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));
         robotsWinningNeuronNeigboursErrorValues.append(gaData_module.getWinningNeuronNeigboursQuantizationErrorOverTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));


      #------ neuron quantization error plotting
      config.outputDirectoryPath = OUTPUT_DIRECTORY + "/" + analysisName;
      visualization.plotNetworkQuantizationErrors(config, envChangedTimes, networkTrainingTimes, robotsWinningNeuronErrorValues, robotsWinningNeuronNeigboursErrorValues, PARAMETER_SET_LABELS);

#------------------------------------
# NETWORK ROBUSTNESS - PARAMETERS
#------------------------------------
if (DO_NETWORK_ROBUSTNESS_PARAMETERS):
   config = resetPlotConfig()
   OUTPUT_DIRECTORY = MAIN_OUTPUT_DIRECTORY + "/networkRobustness_parameters";
   #config.numOfRuns = 4;
   config.style.markers = ['cs','bs','rs'];
   config.style.lineStyles = ['-','-','--'];

   for analysisName in ["corridor"]:
      networkDirectories = getSelectedNetworkDirectories(analysisName);

      #------ habituation threshold training time
      initialNetworkTrainingTimeData = [[] for d in range(len(networkDirectories))]
      averagePassLengths = [[] for d in range(len(networkDirectories))];

      PARAM_SHORTCUT = "ht";
      PARAM_NAME = constants.TEXT_HABITUATION_THRESHOLD;
      PARAM_VALS = [0.00, 0.05, 0.10, 0.15, 0.20, 0.25, 0.30, 0.40, 0.50, 0.60, 0.70, 0.80, 0.90, 1.00];
      PARAM_FORMAT = "{:.2f}";
      config.outputDirectoryPath = OUTPUT_DIRECTORY + "/" + analysisName + "_" + PARAM_SHORTCUT + "_";

      for d in range(len(networkDirectories)):
         dir = networkDirectories[d];
         for p in range(len(PARAM_VALS)):
            paramVal = PARAM_VALS[p];
            config.dataDirectoryPath = paths.BASE_FILE_PATH_DATA + "/" + (dir.replace("x2","x1")).replace("_rms20","") + ("_{}" + PARAM_FORMAT).format(PARAM_SHORTCUT, paramVal);

            envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
            averagePassLength, maxPassLength = helpers.calculatePassLength(config);
            averagePassLengths[d].append(averagePassLength);
            noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);

            initialNetworkTrainingTimeData[d].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));

      config.style.numOfLegendColumns = 3;
      config.style.legendPosition = crGraphStyle.LEGEND_POSITION.UPPER_RIGHT;
      visualization.plotNetworkTrainingTimeByPass(config, initialNetworkTrainingTimeData, averagePassLengths, [[0 for v in range(len(PARAM_VALS))] for d in range(len(networkDirectories))], PARAMETER_SET_LABELS, PARAM_NAME, PARAM_VALS);

      #------ learning rate training time
      initialNetworkTrainingTimeData = [[] for d in range(len(networkDirectories))]
      averagePassLengths = [[] for d in range(len(networkDirectories))];

      PARAM_SHORTCUT = "lr";
      PARAM_NAME = constants.TEXT_LEARNING_RATE;
      PARAM_VALS = [0.00, 0.05, 0.10, 0.15, 0.20, 0.25, 0.30, 0.40, 0.50, 0.60, 0.70, 0.80, 0.90, 1.00];
      PARAM_FORMAT = "{:.2f}";
      config.outputDirectoryPath = OUTPUT_DIRECTORY + "/" + analysisName + "_" + PARAM_SHORTCUT + "_";

      for d in range(len(networkDirectories)):
         dir = networkDirectories[d];
         for p in range(len(PARAM_VALS)):
            paramVal = PARAM_VALS[p];
            config.dataDirectoryPath = paths.BASE_FILE_PATH_DATA + "/" + (dir.replace("x2","x1")).replace("_rms20","") + ("_{}" + PARAM_FORMAT).format(PARAM_SHORTCUT, paramVal);

            envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
            averagePassLength, maxPassLength = helpers.calculatePassLength(config);
            averagePassLengths[d].append(averagePassLength);
            noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);

            initialNetworkTrainingTimeData[d].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));

      config.style.legendPosition = crGraphStyle.LEGEND_POSITION.NONE;
      visualization.plotNetworkTrainingTimeByPass(config, initialNetworkTrainingTimeData, averagePassLengths, [[0 for v in range(len(PARAM_VALS))] for d in range(len(networkDirectories))], PARAMETER_SET_LABELS, PARAM_NAME, PARAM_VALS);
      config.style.legendPosition = crGraphStyle.LEGEND_POSITION.BEST;


#------------------------------------
# NETWORK ROBUSTNESS - ROBOT SPEED
#------------------------------------
if (DO_NETWORK_ROBUSTNESS_ROBOT_SPEED):
   config = resetPlotConfig()
   OUTPUT_DIRECTORY = MAIN_OUTPUT_DIRECTORY + "/networkRobustness_robotSpeed";
   #config.numOfRuns = 4;
   config.style.markers = ['cs','bs','rs'];
   config.style.lineStyles = ['-','-','--'];

   for analysisName in ["corridor","room"]:
      networkDirectories = getSelectedNetworkDirectories(analysisName);

      initialNetworkTrainingTimeData = [[] for d in range(len(networkDirectories))]
      numOfNoveltyTruePositivesData = [[] for d in range(len(networkDirectories))];
      averagePassLengths = [[] for d in range(len(networkDirectories))];

      PARAM_SHORTCUT = "rms";
      PARAM_NAME = constants.TEXT_MAX_ROBOT_SPEED;
      PARAM_VALS = [5, 10, 15, 20, 25, 30, 40, 60, 80];
      PARAM_FORMAT = "{}";
      config.outputDirectoryPath = OUTPUT_DIRECTORY + "/" + analysisName + "_" + PARAM_SHORTCUT + "_";

      for d in range(len(networkDirectories)):
         dir = networkDirectories[d];
         for p in range(len(PARAM_VALS)):
            paramVal = PARAM_VALS[p];
            config.dataDirectoryPath = paths.BASE_FILE_PATH_DATA + "/" + dir.replace("_rms20","") + ("_{}" + PARAM_FORMAT).format(PARAM_SHORTCUT, paramVal);

            envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
            averagePassLength, maxPassLength = helpers.calculatePassLength(config);
            averagePassLengths[d].append(averagePassLength);
            noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);

            initialNetworkTrainingTimeData[d].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));
            numOfNoveltyTruePositivesData[d].append(gaData_module.getNumOfNoveltyTruePositives(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots,config.noveltyValueThreshold, config.robotNoveltyDetectionRange, constants.ENV_CHANGED_EVENT_TYPE));

      #------ plot network training time
      config.style.numOfLegendColumns = 3;
      if ("corridor" in analysisName):
         config.style.legendPosition = crGraphStyle.LEGEND_POSITION.UPPER_RIGHT;
      else:
         config.style.legendPosition = crGraphStyle.LEGEND_POSITION.NONE;
      visualization.plotNetworkTrainingTimeByPass(config, initialNetworkTrainingTimeData, averagePassLengths, [[0 for v in range(len(PARAM_VALS))] for d in range(len(networkDirectories))], PARAMETER_SET_LABELS, PARAM_NAME, PARAM_VALS);

      #------ plot true positives in corridor
      if ("corridor" in analysisName):
         config.style.legendPosition = crGraphStyle.LEGEND_POSITION.LOWER_RIGHT;
         visualization.plotNoveltyReporting(config, numOfNoveltyTruePositivesData, "true positives", PARAMETER_SET_LABELS, PARAM_NAME, PARAM_VALS, convertValuesToPercentages_=True);



#------------------------------------
# NETWORK ROBUSTNESS - NOISE
#------------------------------------
if (DO_NETWORK_ROBUSTNESS_NOISE):
   config = resetPlotConfig()
   OUTPUT_DIRECTORY = MAIN_OUTPUT_DIRECTORY + "/networkRobustness_sensorNoise";
   #config.numOfRuns = 4;
   config.style.markers = ['cs','bs','rs'];
   config.style.lineStyles = ['-','-','--'];

   for analysisName in ["corridor","room"]:
      networkDirectories = getSelectedNetworkDirectories(analysisName);

      initialNetworkTrainingTimeData = [[] for d in range(len(networkDirectories))]
      numOfNeuronsData = [[] for d in range(len(networkDirectories))];  # each data point is maximum number of neurons a network reached, collated over runs
      averagePassLengths = [[] for d in range(len(networkDirectories))];

      PARAM_SHORTCUT = "bcn";
      PARAM_NAME = constants.TEXT_BLOB_CAMERA_NOIZE;
      PARAM_VALS = [0, 0.01, 0.025, 0.05, 0.10, 0.15, 0.20, 0.30, 0.40];
      #PARAM_VALS = [0,0.40];
      PARAM_FORMAT = "{:.3f}";
      config.outputDirectoryPath = OUTPUT_DIRECTORY + "/" + analysisName + "_" + PARAM_SHORTCUT + "_";

      for d in range(len(networkDirectories)):
         dir = networkDirectories[d];
         for p in range(len(PARAM_VALS)):
            paramVal = PARAM_VALS[p];

            if (paramVal == 0):
               #-- resuse data from robot speed experiments, where noize was 0
               config.dataDirectoryPath = paths.BASE_FILE_PATH_DATA + "/" + dir;
            else:
               config.dataDirectoryPath = paths.BASE_FILE_PATH_DATA + "/" + dir + ("_{}" + PARAM_FORMAT).format(PARAM_SHORTCUT, paramVal);

            envChangedTimesData = gaData_module.getEventTimes(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, 1);
            averagePassLength, maxPassLength = helpers.calculatePassLength(config);
            averagePassLengths[d].append(averagePassLength);
            noveltyOccurrenceMinTime = (int)(envChangedTimesData[0][0]);

            initialNetworkTrainingTimeData[d].append(gaData_module.getNetworkTrainingTime(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots, config.noveltyValueThreshold, 0, noveltyOccurrenceMinTime));
            numOfNeuronsData[d].append(gaData_module.getNumOfNeurons(config.dataDirectoryPath, config.numOfRuns, config.numOfRobots));

      #------ plot network training time
      config.style.numOfLegendColumns = 3;
      if ("corridor" in analysisName):
         config.style.legendPosition = crGraphStyle.LEGEND_POSITION.LOWER_RIGHT;
      else:
         config.style.legendPosition = crGraphStyle.LEGEND_POSITION.NONE;
      visualization.plotNetworkTrainingTimeByPass(config, initialNetworkTrainingTimeData, averagePassLengths, [[0 for v in range(len(PARAM_VALS))] for d in range(len(networkDirectories))], PARAMETER_SET_LABELS, PARAM_NAME, PARAM_VALS, yMax_=40);

      #------ plot network size in corridor
      if ("corridor" in analysisName):
         config.style.legendPosition = crGraphStyle.LEGEND_POSITION.UPPER_LEFT;
         brokenPlotYMax = 16000;
         brokenPlotYCoord = 2000;

         visualization.plotNNMetric(config, numOfNeuronsData, "numOfNeurons", constants.TEXT_MAX_NUM_OF_NEURONS, PARAMETER_SET_LABELS, PARAM_NAME, PARAM_VALS, brokenLinePlot_yMax_=brokenPlotYMax,brokenLinePlot_yCoord=brokenPlotYCoord);



