####################################################################
##
## Common functions for visualising data, usually obtained in files with
## s[N]_m[N] prefix.
##
## Author: L. Pitonakova (http://lenkaspace.net)
## License: GNU General Public License. Please credit me when using my work.
##
####################################################################

import copy;
import numpy;

from pyCreeper import crGraphStyle;
from pyCreeper import crGraphs;
from pyCreeper import crData;

import helpers;
import constants;

class PlotConfig(object):
   """
   A configuration object for plots. Contains some constants and styles.
   """
   numOfRobots = 1;
   numOfRuns = 50;
   outputDirectoryPath = "";
   robotNoveltyDetectionRange = 1.05;   # max distance of robot to novelty object at which robot novelty value is examined
   noveltyValueThreshold = 0.9;        # threshold above which a novelty value is considered as corresponding to novelty in the world
   numOfPassesToShowInDetail = 3;      # for how many passes are time series graphs created
   totalExperimentTime = 2000          # in seconds, like in the argos settings file

   def __init__(self):
      self.resetStyle();


   def resetStyle(self):
      self.style = crGraphStyle.crGraphStyle();
      self.style.markers = ['b-', 'c-', 'k-', 'm-', 'g-', 'r-', 'b--', 'c--', 'k--', 'm--', 'g--', 'r--' ];
      self.style.lineStyles = ['-', '-', '-', '-', '-', '-','--', '--', '--', '--', '--','--'];
      self.style.figureSize = (20, 6)
      self.style.legendFontSize = 24;
      self.style.labelFontSize = 24;



#=============================================================================
#                   SINGLE MEASURE PLOTS
#=============================================================================
def plotNetworkTrainingTime(config_, networkTrainingTimeData_, startTime_, endTime_, legendLabels_, xLabel_, xTickLabels_):
   """
   Plot a variable on x-axis and network training time on y-axis.
   :param config_: PlotConfig instance
   :param networkTrainingTimeData_: A list where 0th dimension is different groups (plot colors), 1st dimension is variable shown on the x-axis.
      Each element contains a list returned by gaData_module.getNetworkTrainingTime
   :param startTime_: Start time in simulation ticks (10 per second)
   :param endTime_: End time in simulation ticks (10 per second)
   :param legendLabels_: Labels for the groups (plot colors)
   :param xLabel_: Label for the x-axis (e.g. parameter name)
   :param xTickLabels_: Values on the x-axis
   """
   #config_.style.legendPosition = crGraphStyle.LEGEND_POSITION.UPPER_RIGHT;
   config_.style.xLabelPadding = 0;
   config_.style.figureSize = (12,5);
   config_.style.yOffset = 0.01;
   config_.style.xOffset = 0.025;
   crGraphs.setStyle(config_.style);

   #-- the training time should reflect time since startTime_, while networkTrainingTimeData_ contains absolute times. Adjust and reformat.

   data = copy.deepcopy(networkTrainingTimeData_);
   for i in range(len(data)):
      for j in range(len(data[i])):
         for k in range(len(data[i][j])):
            data[i][j][k] -= startTime_;      # subtract start time
            #-- convert time ticks to minutes
            data[i][j][k] /= 600;


   #-- render
   filePath = config_.outputDirectoryPath + "networkTrainingTime_t" + str(startTime_) + "-" + str(endTime_) + ".png";

   #-- draw horizontal line for end time
   fig = None;
   endTimeInMins = 10000;
   if (startTime_ == 0):
      endTimeInMins = endTime_ / 600;
      lineData = [endTimeInMins for x in range(len(xTickLabels_))];
      origMarkers = config_.style.markers;
      origLineStyles = config_.style.lineStyles;
      config_.style.lineStyles=['--'];
      config_.style.markers = ['k ']
      crGraphs.setStyle(config_.style);
      #fig = crGraphs.createLinePlot(networkTrainingTimeData_, showBoxPlots_=True, yLabel_= "Network training time\n[minutes]", xLabel_=xLabel_, xTickLabels_=xTickLabels_, filePath_=filePath, yMin_=0, yMax_=100);

      #fig = crGraphs.createLinePlot(networkTrainingTimeData_, showBoxPlots_=True, renderFigure_=False, xLabel_=xLabel_, xTickLabels_=xTickLabels_, yMin_=0, yMax_=100);
      config_.style.markers = origMarkers;
      config_.style.lineStyles = origLineStyles;

   if (endTime_ == "end"):
      endTimeInMins = (config_.totalExperimentTime*10 - startTime_) / 600;

   yMax = endTimeInMins;
   crGraphs.setStyle(config_.style);
   yLabel = constants.TEXT_TRAINING_TIME;
   crGraphs.createLinePlot(data, showBoxPlots_=True, yLabel_= yLabel, xLabel_=xLabel_, xTickLabels_=xTickLabels_, legendLabels_=legendLabels_, filePath_=filePath, yMin_=0, yMax_=yMax, figure_=fig);


def plotNetworkTrainingTimeByPass(config_, networkTrainingTimeData_, averagePassLengths_, startTimes_, legendLabels_, xLabel_, xTickLabels_, yMax_=20, startTimeLabel_="0", legendTitle_="", showBoxPlots_=True):
   """
   Plot a variable on x-axis and network training time, translated to number of passes through the environment, on y-axis.
   :param config_: PlotConfig instance
   :param networkTrainingTimeData_: A list where 0th dimension is different groups (plot colors), 1st dimension is variable shown on the x-axis.
      Each element contains a list returned by gaData_module.getNetworkTrainingTime
   :param averagePassLengths_: A 2D array where 0th dimension corresponds to colors on the plot, 1st dimension to values on the x axis.
      Each element is average length of a pass through environment.
   :param startTimes_: A 2D array where 0th dimension corresponds to colors on the plot, 1st dimension to values on the x axis.
      Each element is start time in simulation ticks (10 per second)
   :param legendLabels_: Labels for the groups (plot colors)
   :param xLabel_: Label for the x-axis (e.g. parameter name)
   :param xTickLabels_: Values on the x-axis
   :param yMax_: (optional) Maximum value shown on the y-axis
   :param startTimeLabel_: (optional) Common start time label that is added to the plot file name
   :param legendTitle_: (optional) Title of the legend
   :param showBoxPlots_: (optional) If true, box plots are shown. If false, only medians are shown.
   :return:
   """

   #config_.style.legendPosition = crGraphStyle.LEGEND_POSITION.UPPER_RIGHT;
   config_.style.xLabelPadding = 8;
   config_.style.figureSize = (12,4.5);
   config_.style.yOffset = -0.038;
   config_.style.xOffset = 0.025;
   crGraphs.setStyle(config_.style);

   #-- the training time should reflect time since startTime_, while networkTrainingTimeData_ contains absolute times. Adjust and reformat.
   data = copy.deepcopy(networkTrainingTimeData_);
   for i in range(len(data)):
      for j in range(len(data[i])):
         for k in range(len(data[i][j])):
            data[i][j][k] -= startTimes_[i][j];      # subtract start time
            #-- convert time ticks to pass number, based on pass length
            data[i][j][k] /= averagePassLengths_[i][j];

            #-- post-process the num of passes to max of 20, so that graphs of the same scales can be shown for before and after training
            if (data[i][j][k] > yMax_):
               data[i][j][k] = yMax_;

   #-- render
   filePath = config_.outputDirectoryPath + "networkTrainingTimeAsPass_" + str(startTimeLabel_) + ".png";

   yLabel = constants.TEXT_TRAINING_TIME_PASSES;
   crGraphs.setStyle(config_.style);
   crGraphs.createLinePlot(data, showBoxPlots_=showBoxPlots_, yLabel_= yLabel, xLabel_=xLabel_, xTickLabels_=xTickLabels_, legendLabels_=legendLabels_, filePath_=filePath, yMin_=0, yMax_=yMax_, legendTitle_=legendTitle_);
   #crGraphs.createLinePlot(data, showAverages_=True, showConfidenceIntervals_=True, yLabel_= yLabel, xLabel_=xLabel_, xTickLabels_=xTickLabels_, legendLabels_=legendLabels_, filePath_=filePath, yMin_=0, yMax_=yMax_, legendTitle_=legendTitle_);

   #-- output text file with medians and averages
   filePath = config_.outputDirectoryPath + "networkTrainingTimeAsPass_" + str(startTimeLabel_);
   printDataStats(data, legendTitle_, legendLabels_, xLabel_, xTickLabels_, filePath);



def plotNumOfRunsNovelObjectReported(config_, numOfRunsNovelObjectReportedRawData_, legendLabels_, xLabel_, xTickLabels_):
   """
   Plot line plot where x-axis and colors are parameters, y-axis is number of runs.
   N plots are created, one per pass through environment.

   :param config_: PlotConfig instance
   :param numOfRunsNovelObjectReportedRawData_: A list where 0th dimension is different groups (plot colors), 1st dimension is variable shown on the x-axis.
      Each element contains a list returned by gaData_module.getNumOfRunsNovelObjectReported
   :param legendLabels_: Labels for the groups (plot colors)
   :param xLabel_: Label for the x-axis (e.g. parameter name)
   :param xTickLabels_: Values on the x-axis
   """
   config_.style.xLabelPadding = 0;
   config_.style.figureSize = (12,5);
   config_.style.yOffset = 0.01;
   config_.style.xOffset = 0.025;
   crGraphs.setStyle(config_.style);

   #-- the numOfRunsNovelObjectReportedRawData_ is indexed by [0] param 1 (color), [1] param 2 (x-axis), [3] pass no.
   #   change this to [0] pass no [1] param 1 (color), [2] param 2 (x-axis)
   arrayToDisplay = [];
   for passNo in range(len(numOfRunsNovelObjectReportedRawData_[0][0])):
      arrLvl1 = []
      for x in range(len(numOfRunsNovelObjectReportedRawData_)):
         arrLvl2 = []
         for y in range(len(numOfRunsNovelObjectReportedRawData_[x])):
            arrLvl2.append((numOfRunsNovelObjectReportedRawData_[x][y][passNo] / config_.numOfRuns) * 100)  # also convert to percentage
         arrLvl1.append(arrLvl2)
      arrayToDisplay.append(arrLvl1)

   #print(arrayToDisplay)

   #-- plot
   outputDirectoryPath = config_.outputDirectoryPath + "/novelObjectReporting_metrics/";
   helpers.createFolder(outputDirectoryPath);
   for passNo in range(config_.numOfPassesToShowInDetail):
   #for passNo in range(len(arrayToDisplay)): #
      filePath = outputDirectoryPath + "runsNovelObjectReported_pass" + str(passNo) + ".png";
      crGraphs.createLinePlot(arrayToDisplay[passNo], yLabel_="Percentage of runs\nnovelty reported in pass " + (str(passNo)), legendLabels_=legendLabels_, xLabel_=xLabel_, xTickLabels_=xTickLabels_, filePath_=filePath, yMin_=0, yMax_=100);



def plotNoveltyReporting(config_, noveltyReportingData_, dataTitle_, legendLabels_, xLabel_, xTickLabels_, legendTitle_ = "",
      yMax_=crGraphs.INVALID_VALUE, brokenLinePlot_yMax_=crGraphs.INVALID_VALUE, brokenLinePlot_yCoord=crGraphs.INVALID_VALUE, brokenLinePlot_heightPercentage_=60,
      showAverages_ = True, convertValuesToPercentages_ = False):
   """
   Plot box plot where x-axis and colors are parameters, y-axis is number of reports of a certain type, e.g. of novely true positives.
   N plots are created, one per pass through the environment.
   :param config_: PlotConfig instance
   :param noveltyReportingData_:  A list where 0th dimension is different groups (plot colors), 1st dimension is variable shown on the x-axis.
      Each element contains a list returned by one of gaData_module.getNumOfNoveltyTruePositives, gaData_module.getNumOfNoveltyFalseNegatives, gaData_module.getNumOfNoveltyFalsePositives
   :param dataTitle_: Title of the type of data, e.g. "true positives". Will be added to file name and y-axis label
   :param legendLabels_: Labels for the groups (plot colors)
   :param xLabel_: Label for the x-axis (e.g. parameter name)
   :param xTickLabels_: Values on the x-axis
   :param legendTitle_: (optional) Title of the legend
   :param yMax_: (optional) Maximum value shown on the y-axis
   :param brokenLinePlot_yMax_: (optional) Maximum value shown on the y-axis of a broken plot - see pyCreeper.crGraphs.getBrokenLinePlotParameters
   :param brokenLinePlot_yCoord: (optional): Y-axis value at which the broken line will be drawn - see pyCreeper.crGraphs.getBrokenLinePlotParameters
   :param brokenLinePlot_heightPercentage_: (optional) The percentage (20-80) of the plot's height at which the broken line will be drawn - see pyCreeper.crGraphs.getBrokenLinePlotParameters
   :param showAverages_: (optional) If true, averages are shown. If false, box plots are shown
   :param convertValuesToPercentages_: (optional) If true, y-axis values are multiplied by 100.
   """


   config_.style.xLabelPadding = 8;
   config_.style.figureSize = (12,4);
   config_.style.yOffset = -0.038;
   config_.style.xOffset = 0.025;
   crGraphs.setStyle(config_.style);

   # -- the numOfRunsNovelObjectReportedRawData_ is indexed by [0] param 1 (color), [1] param 2 (x-axis), [3] pass no.
   #   change this to [0] pass no [1] param 1 (color), [2] param 2 (x-axis)
   arrayToDisplay = [];
   for passNo in range(len(noveltyReportingData_[0][0])):
      arrLvl1 = []
      for x in range(len(noveltyReportingData_)):
         arrLvl2 = []
         for y in range(len(noveltyReportingData_[x])):
            if (convertValuesToPercentages_):
               #-- convert to percentage - only use when there is only a single novelty to report in each pass
               arrLvl2.append([noveltyReportingData_[x][y][passNo][v]*100 for v in range(len(noveltyReportingData_[x][y][passNo]))]);
            else:
               arrLvl2.append([noveltyReportingData_[x][y][passNo][v] for v in range(len(noveltyReportingData_[x][y][passNo]))]);
         arrLvl1.append(arrLvl2)
      arrayToDisplay.append(arrLvl1)


   # -- plot

   if ("true positives" in dataTitle_):
      yMax_ = 1.0
      if (convertValuesToPercentages_):
         yMax_ = 100;

   outputDirectoryPath = config_.outputDirectoryPath + "/novelObjectReporting_metrics/";
   helpers.createFolder(outputDirectoryPath);

   showAverages = showAverages_;
   showBoxPlots = False;
   if (showAverages == False):
      showBoxPlots = True;

   for passNo in range(config_.numOfPassesToShowInDetail):
      yLabel = "Number of " + dataTitle_ + "\nin pass " + (str(passNo));
      #-- special y labels
      if (passNo == 0):
         if ("true positives" in dataTitle_):
            yLabel = "% of runs in which   \nnovelty was reported   ";
         elif ("false positives" in dataTitle_):
            yLabel = "Num. of false positives  \nreported after training  ";
      #--
      if (brokenLinePlot_yMax_ == crGraphStyle.INVALID_VALUE):
         #-- draw normal plot
         filePath = outputDirectoryPath + dataTitle_ + "_pass" + str(passNo) + ".png";
         crGraphs.createLinePlot(arrayToDisplay[passNo], showAverages_=showAverages, showBoxPlots_=showBoxPlots, yLabel_=yLabel, legendLabels_=legendLabels_, xLabel_=xLabel_, xTickLabels_=xTickLabels_, filePath_=filePath, yMin_=0, yMax_=yMax_, legendTitle_=legendTitle_);
      else:
          #-- broken line plot params
         yMinStandardPlot = 0; # y min of a plot that would normally be drawn for all data
         yMaxStandardPlot = brokenLinePlot_yMax_; # y max of a plot that would normally be drawn for all data
         brokenLineYCoordinate = brokenLinePlot_yCoord; # y value at which broken line will be drawn. This value separates the two "zoomedIn" and "zoomedOut" plots
         brokenLineHeightPercentage = brokenLinePlot_heightPercentage_; # percentage of the plot height at which the broken line will be drawn

         yMinBottomPlot, yMaxBottomPlot, yMinTopPlot, yMaxTopPlot = crGraphs.getBrokenLinePlotParameters(yMinStandardPlot, brokenLineYCoordinate, yMaxStandardPlot, brokenLineHeightPercentage);

         #-- draw broken line plot (2 figs)
         filePath = outputDirectoryPath + dataTitle_ + "_pass" + str(passNo) + "_bottom.png";
         crGraphs.createLinePlot(arrayToDisplay[passNo], showAverages_=showAverages, showBoxPlots_=showBoxPlots, yLabel_=yLabel, legendLabels_=legendLabels_, xLabel_=xLabel_, xTickLabels_=xTickLabels_, filePath_=filePath, yMin_=yMinBottomPlot, yMax_=yMaxBottomPlot, legendTitle_=legendTitle_);
         filePath = outputDirectoryPath + dataTitle_ + "_pass" + str(passNo) + "_top.png";
         crGraphs.createLinePlot(arrayToDisplay[passNo], showAverages_=showAverages, showBoxPlots_=showBoxPlots, yLabel_=yLabel, legendLabels_=legendLabels_, xLabel_=xLabel_, xTickLabels_=xTickLabels_, filePath_=filePath, yMin_=yMinTopPlot, yMax_=yMaxTopPlot, legendTitle_=legendTitle_);


def plotNoveltyReportingByPass(config_, data_, dataTitle_, legendLabels_, maxPassNo_=10, yMax_=crGraphs.INVALID_VALUE, brokenLinePlot_yMax_=crGraphs.INVALID_VALUE, brokenLinePlot_yCoord=crGraphs.INVALID_VALUE):
   """
   Plot box plot where x-axis is pass number after novelty has occurred, colors are different parameter sets.
   The data_ should have its 0th dimension the different parameter sets, the 1st dimension passes, the 2nd dimension invidiual runs.

   :param config_: PlotConfig instance
   :param data_: A list where 0th dimension the different parameter sets (colors).
      Each element contains a list returned by one of gaData_module.getNumOfNoveltyTruePositives, gaData_module.getNumOfNoveltyFalseNegatives, gaData_module.getNumOfNoveltyFalsePositives
   :param dataTitle_: Title of the type of data, e.g. "true positives". Will be added to file name and y-axis label
   :param legendLabels_: Labels for the groups (plot colors)
   :param maxPassNo_: (optional) Maximum number of passes to show
   :param yMax_: (optional) Maximum value shown on the y-axis
   :param brokenLinePlot_yMax_: (optional) Maximum value shown on the y-axis of a broken plot - see pyCreeper.crGraphs.getBrokenLinePlotParameters
   :param brokenLinePlot_yCoord: (optional): Y-axis value at which the broken line will be drawn - see pyCreeper.crGraphs.getBrokenLinePlotParameters
   """


   config_.style.xLabelPadding = 0;
   config_.style.figureSize = (12,5);
   config_.style.yOffset = 0.01;
   config_.style.xOffset = 0.025;
   crGraphs.setStyle(config_.style);

   # -- plot
   outputDirectoryPath = config_.outputDirectoryPath + "/novelObjectReporting_metrics/";
   helpers.createFolder(outputDirectoryPath);


   maxPassNo = maxPassNo_;
   if (maxPassNo == crGraphs.INVALID_VALUE):
      maxPassNo = len(data_[0]);

   xTickLabels = [x for x in range(maxPassNo)];


   arrayToDisplay = [];
   for x in range(len(data_)):
      arrLvl1 = []
      for passNo in range(maxPassNo):
         arrLvl1.append([data_[x][passNo][v] for v in range(len(data_[x][passNo]))]);
      arrayToDisplay.append(arrLvl1)

   if (brokenLinePlot_yMax_ == crGraphStyle.INVALID_VALUE):
      #-- draw normal plot
      filePath = outputDirectoryPath + dataTitle_ + "_selectedParams.png";
      crGraphs.createLinePlot(arrayToDisplay, showAverages_=True, yLabel_= "Number of " + dataTitle_, legendLabels_=legendLabels_, xLabel_="Pass", xTickLabels_=xTickLabels, filePath_=filePath, yMin_=0, yMax_=yMax_ );
   else:
      #-- broken line plot params
      yMinStandardPlot = 0; # y min of a plot that would normally be drawn for all data
      yMaxStandardPlot = brokenLinePlot_yMax_; # y max of a plot that would normally be drawn for all data
      brokenLineYCoordinate = brokenLinePlot_yCoord; # y value at which broken line will be drawn. This value separates the two "zoomedIn" and "zoomedOut" plots
      brokenLineHeightPercentage = 60; # percentage of the plot height at which the broken line will be drawn

      yMinBottomPlot, yMaxBottomPlot, yMinTopPlot, yMaxTopPlot = crGraphs.getBrokenLinePlotParameters(yMinStandardPlot, brokenLineYCoordinate, yMaxStandardPlot, brokenLineHeightPercentage);

      #-- draw broken line plot (2 figs)
      filePath = outputDirectoryPath + dataTitle_ + "_selectedParams_bottom.png";
      crGraphs.createLinePlot(arrayToDisplay, showAverages_=True,  yLabel_= "Number of " + dataTitle_, legendLabels_=legendLabels_, xLabel_="Pass", xTickLabels_=xTickLabels, filePath_=filePath, yMin_=yMinBottomPlot, yMax_=yMaxBottomPlot);
      filePath = outputDirectoryPath + dataTitle_ + "_selectedParams_top.png";
      crGraphs.createLinePlot(arrayToDisplay, showAverages_=True,  yLabel_= "Number of " + dataTitle_, legendLabels_=legendLabels_, xLabel_="Pass", xTickLabels_=xTickLabels, filePath_=filePath, yMin_=yMinTopPlot, yMax_=yMaxTopPlot);



   #-- draw normal plot


def plotNNMetric(config_, rawData_, fileName_, yLabel_, legendLabels_, xLabel_, xTickLabels_, yMin_=0, yMax_=crGraphs.INVALID_VALUE, brokenLinePlot_yMax_=crGraphs.INVALID_VALUE, brokenLinePlot_yCoord=crGraphs.INVALID_VALUE, legendTitle_=""):
   """
   Plot box plot where x-axis and colors are parameters, y-axis is a metric.
   :param config_: PlotConfig instance
   :param rawData_:  A list where 0th dimension is different groups (plot colors), 1st dimension is variable shown on x-axis.
      Each element contains a list returned by a gaData_module single metric function, such as gaData_module.getNumOfNeurons
   :param fileName_: File name of the plot
   :param yLabel_: Label for the y-axis
   :param legendLabels_: Labels for the groups (plot colors)
   :param xLabel_: Label for the x-axis (e.g. parameter name)
   :param xTickLabels_: Values on the x-axis
   :param yMin_: (optional) Minimum value shown on the y-axis
   :param yMax_: (optional) Maximum value shown on the y-axis
   :param brokenLinePlot_yMax_: (optional) Maximum value shown on the y-axis of a broken plot - see pyCreeper.crGraphs.getBrokenLinePlotParameters
   :param brokenLinePlot_yCoord: (optional): Y-axis value at which the broken line will be drawn - see pyCreeper.crGraphs.getBrokenLinePlotParameters
   :param legendTitle_: (optional) Title of the legend
   """

   config_.style.xLabelPadding = 12;
   config_.style.figureSize = (12,4.8);
   config_.style.yOffset = -0.030;
   config_.style.xOffset = 0.025;
   crGraphs.setStyle(config_.style);


   #-- plot
   outputDirectoryPath = config_.outputDirectoryPath + "/neuralNetwork_metrics/";
   helpers.createFolder(outputDirectoryPath);

   if (brokenLinePlot_yMax_ == crGraphStyle.INVALID_VALUE):
      #-- draw normal plot
      filePath = outputDirectoryPath + fileName_ + ".png"
      crGraphs.createLinePlot(rawData_, showBoxPlots_=True, yLabel_=yLabel_, legendLabels_=legendLabels_, xLabel_=xLabel_, xTickLabels_=xTickLabels_, filePath_=filePath, yMin_=yMin_, yMax_=yMax_, legendTitle_=legendTitle_);
   else:
      #-- broken line plot params
      yMinStandardPlot = yMin_; # y min of a plot that would normally be drawn for all data
      yMaxStandardPlot = brokenLinePlot_yMax_; # y max of a plot that would normally be drawn for all data
      brokenLineYCoordinate = brokenLinePlot_yCoord; # y value at which broken line will be drawn. This value separates the two "zoomedIn" and "zoomedOut" plots
      brokenLineHeightPercentage = 60; # percentage of the plot height at which the broken line will be drawn

      yMinBottomPlot, yMaxBottomPlot, yMinTopPlot, yMaxTopPlot = crGraphs.getBrokenLinePlotParameters(yMinStandardPlot, brokenLineYCoordinate, yMaxStandardPlot, brokenLineHeightPercentage);

      #-- draw broken line plot (2 figs)
      filePath = outputDirectoryPath + fileName_ + "_bottom.png";
      crGraphs.createLinePlot(rawData_, showBoxPlots_=True,  yLabel_=yLabel_, legendLabels_=legendLabels_, xLabel_=xLabel_, xTickLabels_=xTickLabels_, filePath_=filePath, yMin_=yMinBottomPlot, yMax_=yMaxBottomPlot, legendTitle_=legendTitle_);
      filePath = outputDirectoryPath + fileName_ + "_top.png";
      crGraphs.createLinePlot(rawData_, showBoxPlots_=True,  yLabel_=yLabel_, legendLabels_=legendLabels_, xLabel_=xLabel_, xTickLabels_=xTickLabels_, filePath_=filePath, yMin_=yMinTopPlot, yMax_=yMaxTopPlot, legendTitle_=legendTitle_);

   #-- output text file with medians and averages
   printDataStats(rawData_, legendTitle_, legendLabels_, xLabel_, xTickLabels_, outputDirectoryPath + fileName_);



def plotNNUsageHistogram(config_: PlotConfig, data_, fileName_, firstNBins_ = crGraphs.INVALID_VALUE, lastNBins_=crGraphs.INVALID_VALUE, yMax_ = crGraphs.INVALID_VALUE):
   """
   Plot a histogram of neuron fire counts, where x-axis is fire counts and y-axis is percentage of neurons with that fire count.
   Also creates a text file that shows the total average fire count, considering all neurons.
   :param config_: PlotConfig instance
   :param data_: A list returned by gaData_module.getNeuronUsageAsBins
   :param fileName_: File name of the plot
   :param firstNBins_: (optional) Number of bins with the lowest count to show only
   :param lastNBins_: (optional) Number of bins with the highest count to show only
   :param yMax_: (optional) Maximum value shown on the y-axis
   """
   #-- plot
   crGraphs.setStyle(config_.style);
   barWidth = 0.8;
   for robotId in range(len(data_)):

      if (firstNBins_ != crGraphs.INVALID_VALUE):
         #-- plot graph for first N
         indexStart = 0;
         indexEnd = firstNBins_;
         filePath = config_.outputDirectoryPath + "_" + fileName_ + "_robot" + str(robotId) + "_first" + str(firstNBins_) + ".png";
         createBarChart([data_[robotId][indexStart:indexEnd]], yLabel_ = "Percentage of neurons", xLabel_="Fire count", yMin_=0, yMax_ = yMax_, barWidth_=barWidth, filePath_=filePath);

      if (lastNBins_ != crGraphs.INVALID_VALUE):
         #-- plot graph for last N
         indexStart = len(data_[robotId]) - lastNBins_;
         indexEnd = len(data_[robotId]);
         xTickLabels = [len(data_[robotId]) - lastNBins_ + x for x in range(lastNBins_)];
         filePath = config_.outputDirectoryPath + "_" + fileName_ + "_robot" + str(robotId) + "_last" + str(lastNBins_) + ".png";
         createBarChart([data_[robotId][indexStart:indexEnd]], yLabel_ = "Percentage of neurons", xLabel_="Fire count", xTickLabels_=xTickLabels, yMin_=0, yMax_ = yMax_, barWidth_=barWidth, filePath_=filePath);

      if (lastNBins_ == crGraphs.INVALID_VALUE and firstNBins_ == crGraphs.INVALID_VALUE):
         #-- plot graph for all data
         filePath = config_.outputDirectoryPath + "_" + fileName_ + "_robot" + str(robotId) + ".png";
         createBarChart([data_[robotId]], yLabel_ = "Percentage of neurons", xLabel_="Fire count", yMin_=0, yMax_ = yMax_, barWidth_=barWidth, filePath_=filePath);

   #-- consider the whole data set to calculate total average fire count
   outputStr = "";
   for robotId in range(len(data_)):
      outputStr += "Robot " + str(robotId);
      totalFireCounts = 0;
      for i in range(len(data_[robotId])):
         totalFireCounts += i * data_[robotId][i];
      averageFireCount = totalFireCounts / sum(data_[robotId]);
      outputStr += " Average fire count: " + str(averageFireCount)
      outputStr += "\n"

   f = open(config_.outputDirectoryPath + "__stats.txt",'w')
   f.write(outputStr);
   f.close()

#=============================================================================
#                   TIME SERIES PLOTS
#=============================================================================

def plotNeuronUsageTimeSeries(config_, robotsNetworkSizeData_, robotsNumOfUsedNeuronsData_, robotsNumOfUsedNeuronsAfterTrainingData_):
   """
   Plot time on x-axis and percentage of neurons, out of total neurons in the current time step, on y-axis.
   :param config_: PlotConfig instance
   :param robotsNetworkSizeData_: A list returned by gaData_module.getNumOfNeuronsOverTime
   :param robotsNumOfUsedNeuronsData_: A list returned by gaData_module.getNumOfUsedNeuronsOverTimeSince
   :param robotsNumOfUsedNeuronsAfterTrainingData_: A list returned by gaData_module.getNumOfUsedNeuronsOverTimeSince
   """
   robotsNumOfUsedNeurons = [[[0 for v in range(len(robotsNetworkSizeData_[robotId][t]))] for t in range(len(robotsNetworkSizeData_[robotId]))] for robotId in range(len(robotsNetworkSizeData_)) ];
   robotsNumOfUsedNeuronsAfterTraining = [[[0 for v in range(len(robotsNetworkSizeData_[robotId][t]))] for t in range(len(robotsNetworkSizeData_[robotId]))] for robotId in range(len(robotsNetworkSizeData_)) ];


   for robotId in range(len(robotsNetworkSizeData_)):
      for t in range(len(robotsNetworkSizeData_[robotId])):
         for v in range(len(robotsNetworkSizeData_[robotId][t])):
            robotsNumOfUsedNeurons[robotId][t][v] = robotsNumOfUsedNeuronsData_[robotId][t][v] / robotsNetworkSizeData_[robotId][t][v];
            robotsNumOfUsedNeuronsAfterTraining[robotId][t][v] = robotsNumOfUsedNeuronsAfterTrainingData_[robotId][t][v] / robotsNetworkSizeData_[robotId][t][v];

   plotTimeSeries(config_, [robotsNumOfUsedNeurons, robotsNumOfUsedNeuronsAfterTraining],"networkNeuronsUsage_percentages.png","Number of neurons", 1.0, legendLabels_=["Used since the beginning", "Used since change"]);


def plotTimeSeries(config_: PlotConfig, data_, fileName_, yLabel_, yMax_=crGraphs.INVALID_VALUE, timeStart_ = 0, timeEnd_=crGraphs.INVALID_VALUE, legendLabels_=[]):
   """
   Plot time on the x axis and a metric on the y-axis. Average across multiple runs.
   :param config_: PlotConfig instance
   :param data_: A list where 0th dimension is different metrics (plot colors).
      Each element is a lst returned by a gaData_module time series function, e.g. gaData_module.getNumOfNeuronsOverTime
   :param fileName_: File name of the plot
   :param yLabel_: Label for the y-axis
   :param yMax_: (optional) Maximum value shown on the y-axis
   :param timeStart_: Start time in simulation ticks (10 per second)
   :param timeEnd_: End time in simulation ticks (10 per second)
   :param legendLabels_: Labels for the metrics (plot colors)
   """
   robotId = 0; #TODO

   if (timeEnd_ == crGraphs.INVALID_VALUE):
      timeEnd_ = config_.totalExperimentTime*10;


   #-- create array of novelty values for the first N passes
   selectedNPassesData = [];

   for line in range(len(data_)):
      selectedNPassesData.append((data_[line][robotId][(int)(timeStart_):(int)(timeEnd_)]));  # TODO: do this for N robots to show in 1 graph

   #-- get average robot teleport times so they can be shown on the graph x-axis
   averagePassLength, maxPassLength = helpers.calculatePassLength(config_);
   averageRobotTeleportTimes = [i*averagePassLength for i in range(1,40,1)]

   #-- render
   config_.style.yOffset = -0.05
   timeStep = 30;
   if (timeEnd_ - timeStart_ > 10000):
      timeStep = 120;
   elif (timeEnd_ - timeStart_ > 5000):
      timeStep = 60;


   xTickLabelPositions, xTickLabels = helpers.generateXTickPositionsAndLabels_time(data_[0][robotId], timeStart_, -1, timeStep, averageRobotTeleportTimes, nonRegularTimesStringFormat_="\n|") # 10 ticks = 1 second


   counter = 0;
   for x in range(len(xTickLabels)):
      if (xTickLabels[x] == "\n|"):
         counter += 1;
         if (counter == 20):
            xTickLabels[x] = "\n^|^"

   filePath = config_.outputDirectoryPath + "_" + fileName_;
   #filePath = "";

   config_.style.markers = ['b-','c-','m-'];
   crGraphs.setStyle(config_.style);
   crGraphs.createLinePlot(selectedNPassesData, yLabel_=yLabel_, showAverages_=True, averagesIgnoreZeros_=True, xLabel_="Time [seconds]", yMin_=0, yMax_=yMax_, legendLabels_=legendLabels_, xTickLabels_=xTickLabels,xTickLabelPositions_=xTickLabelPositions, filePath_=filePath);
   config_.style.yOffset = 0.0;



def plotNoveltyValueAroundNovelObject(config_: PlotConfig, noveltyValues_, robotsAroundNoveltyTimesData_, robotsNoveltyValueByDistanceToNoveltyData_, distanceRange_, numOfPassesToShow_ = 3):
   """
   Plot distance to novel object on x-axis, novelty value on y axis. Zero-averages across multiple runs.
   N files will be generated, one for each pass around the novel object.

   :param config_: PlotConfig instance
   :param noveltyValues_: A list returned by gaData_module.getNoveltyValues
   :param robotsAroundNoveltyTimesData_: A list returned by gaData_module.getRobotsAroundNoveltyTimes
   :param robotsNoveltyValueByDistanceToNoveltyData_: A list returned by gaData_module.getNoveltyValuesByDistanceToNovelty
   :param distanceRange_: Range in metres from the novel object position to show data for. I.e., x-axis will contain values from -distanceRange_ to distanceRange_
   :param numOfPassesToShow_: (optional) Number of passes to make figures for
   :return:
   """
   crGraphs.setStyle(config_.style);
   averagePassLength, maxPassLength = helpers.calculatePassLength(config_);
   robotId=0 #TODO

   #-- create an array that sorts the robot's novelty data per run, so that each run can be shown with a different color.
   #   this array will only have value for periods when robot was around novel object.
   noveltyPassesData_linePerRun = [];
   numOfRunsToShow = min(config_.numOfRuns,5);
   for runNo in range(numOfRunsToShow):
      noveltyPassesData_linePerRun.append([]);
      for t in range(len(noveltyValues_[robotId])):
         timeStart = robotsAroundNoveltyTimesData_[robotId][runNo][0];  # TODO: needs doing for multiple env changes in 1 run
         timeEnd = robotsAroundNoveltyTimesData_[robotId][runNo][0] + averagePassLength * 2;
         #print(t, timeStart, timeEnd);
         if (t >= timeStart and t <= timeEnd):
            noveltyPassesData_linePerRun[runNo].append(noveltyValues_[robotId][t][runNo]);

   #-- plot average novelty value while robot was in the range of novel object
   #legendLabels = ['Run ' + str(r) for r in range(numOfRunsToShow)];
   #filePath = config_.outputDirectoryPath + "_avgNoveltyValAroundNovelty.png"
   #filePath = ""
   #xTickLabelPositions, xTickLabels = helpers.generateXTickPositionsAndLabels_time(noveltyPassesData_linePerRun[0], 0, -1, 5)
   #crGraphs.createLinePlot(noveltyPassesData_linePerRun, yLabel_="Avg. novelty", showAverages_=False, xLabel_="Time after novelty in range [seconds]", xTickLabels_=xTickLabels,xTickLabelPositions_=xTickLabelPositions, legendLabels_=legendLabels, yMin_=0, yMax_=1, filePath_=filePath);

   # -- plot novelty value by distance from novel object   #TODO: draw something through the middle of the graph to indicate distances that could not be reached?
   numOfPassesToShow = min(numOfPassesToShow_, len(robotsNoveltyValueByDistanceToNoveltyData_[robotId]));
   for passNo in range(numOfPassesToShow):
      filePath = config_.outputDirectoryPath + "_avgNoveltyValByDistanceToNovelty_pass" + str(passNo) + ".png";
      xTickLabelPositions, xTickLabels = helpers.generateXTickPositionsAndLabels(robotsNoveltyValueByDistanceToNoveltyData_[robotId][passNo][0], -distanceRange_, distanceRange_, len(robotsNoveltyValueByDistanceToNoveltyData_[robotId][passNo][0]))
      for i in range(len(xTickLabels)):
         if (i % 10 != 0):
            xTickLabels[i] = ""

      # -- flip the displayed array so that all runs are considered together for 1 data point and averages displayed
      array = crData.getListByFlippingColumnsAndRows(robotsNoveltyValueByDistanceToNoveltyData_[robotId][passNo]);
      crGraphs.createLinePlot([array], yLabel_="Max novelty\nin pass " + str(passNo), showAverages_=True, averagesIgnoreZeros_= True, xLabel_="Distance to novel object [metres]", xTickLabels_=xTickLabels, yMin_=0, yMax_=1, filePath_=filePath);

def plotNetworkQuantizationErrors(config_, envChangedTimes_, networkTrainingTimes_, robotsWinningNeuronErroValues_, robotsWinningNeuronNeighboursErroValues_, parameterSetLabels_):
   """
   Plot different networks (parameter sets) on x-axis and quantization error of winning neuron and its neighbours on y-axis.
   :param config_: PlotConfig instance
   :param envChangedTimes_: A list where 0th dimension is parameter sets.
      Each element contains a list returned by gaData_module.getEventTimes for the environment changed event type
   :param networkTrainingTimes_: A list where 0th dimension is parameter sets.
      Each element contains a list returned by gaData_module.getNetworkTrainingTime
   :param robotsWinningNeuronErroValues_: A list where 0th dimension is parameter sets.
      Each element contains a list returned by gaData_module.getWinningNeuronQuantizationErrorOverTime
   :param robotsWinningNeuronNeighboursErroValues_: A list where 0th dimension is parameter sets.
      Each element contains a list returned by gaData_module.getWinningNeuronNeigboursQuantizationErrorOverTime
   :param parameterSetLabels_: X-axis labels for the different parameter sets.
   """
   robotId = 0;

   winningNeuronErrorValuesAfterTraining = [[] for paramSet in range(len(robotsWinningNeuronErroValues_))];
   winningNeuronNeighboursErrorValuesAfterTraining = [[] for paramSet in range(len(robotsWinningNeuronErroValues_))];

   for paramSet in range(len(robotsWinningNeuronErroValues_)):
      #-- the error values are ordered first by time, then by runno, swap this ordering
      winningNeuronErroValues = crData.getListByFlippingColumnsAndRows(robotsWinningNeuronErroValues_[paramSet][robotId])
      winningNeuronNeighboursErroValues = crData.getListByFlippingColumnsAndRows(robotsWinningNeuronNeighboursErroValues_[paramSet][robotId])

      for runNo in range(config_.numOfRuns):
         networkTrainingDoneTime = (int)(networkTrainingTimes_[paramSet][runNo]);
         envChangedTime = (int)(envChangedTimes_[paramSet][runNo][0]);

         winningNeuronErrorValuesAfterTraining[paramSet].append(crData.getAverageOfAList(winningNeuronErroValues[runNo][networkTrainingDoneTime+1:envChangedTime]));
         winningNeuronNeighboursErrorValuesAfterTraining[paramSet].append(crData.getAverageOfAList(winningNeuronNeighboursErroValues[runNo][networkTrainingDoneTime+1:envChangedTime]));

   #-- plot
   crGraphs.setStyle(config_.style);
   crGraphs.getStyle().figureSize = (5.5,4);
   crGraphs.getStyle().markers = ['bs','ms'];
   crGraphs.getStyle().markerSize = 15;
   crGraphs.getStyle().boxPlotLineWidth = 2;
   crGraphs.getStyle().lineStyles = ['',''];
   crGraphs.getStyle().numOfLegendColumns = 1;
   crGraphs.getStyle().boxPlotWidth = 0.5;
   filePath = config_.outputDirectoryPath + "_quantizationErrors.png";

   yLabel = constants.TEXT_QUANTIZATION_ERROR;
   legendLabels = [] #["Winning neuron", "Neighbours"]; #legend is too much in the way of ths small graph - will have to be added manually
   crGraphs.createLinePlot([winningNeuronErrorValuesAfterTraining, winningNeuronNeighboursErrorValuesAfterTraining], yMin_=0, yMax_=2.0, yLabel_=yLabel, showBoxPlots_=True, legendLabels_=legendLabels, xTickLabels_=parameterSetLabels_, filePath_=filePath);



#=============================================================================
#                   TEXT-BASED
#=============================================================================
def printDataStats(rawData_, dimension0Title_, dimension0Labels_, dimension1Title_, dimension1Labels_, filePathAndPrefix_ ):
   """
   Print data statistics like average and median value in a text file
   :param rawData_: A list where 0th and 1st dimensions are parameters and 2nd dimension are values to take statistics of.
      E.g. data returned by gaData_module.getNetworkTrainingTime
   :param dimension0Title_: Title of the parameter on 0th data dimension. In plots, this is usually legend title
   :param dimension0Labels_: Labels for the parameter on 0th data dimension. In plots, this is usually legend labels
   :param dimension1Title_: Title of the parameter on 1st data dimension. In plots, this is usually the x-axis label
   :param dimension1Labels_: Labels for the parameter on 1st data dimension. In plots, this is usually x-axis labels
   :param filePathAndPrefix_: Full file path to the outputted text file, including the file name without extension
   """
   outputStr = "";
   for i in range(len(rawData_)):
      outputStr += dimension0Title_ + " " + str(dimension0Labels_[i]) + "\n------------------\n\n";
      for j in range(len(rawData_[i])):
         outputStr += dimension1Title_ + " " + str(dimension1Labels_[j]) + ": ";
         outputStr += "Average = " + str(crData.getAverageOfAList(rawData_[i][j])) + " ";
         outputStr += "Median = " + str(crData.getMedianOfAList(rawData_[i][j]));
         outputStr += "\n"
      outputStr += "\n\n"
   f = open(filePathAndPrefix_ + "__stats.txt",'w')
   f.write(outputStr);
   f.close()

#=============================================================================


from pyCreeper import crHelpers;
import scipy;
import math;
import itertools;
import matplotlib.pyplot as plt;

def createBarChart(data_,
                    title_="", xLabel_ = "", yLabel_ = "", xTickLabels_=[], legendLabels_ = [],
                    showConfidenceIntervals_=False, barWidth_ = 0.35,
                    xMin_=crGraphs.INVALID_VALUE, xMax_=crGraphs.INVALID_VALUE, yMin_=crGraphs.INVALID_VALUE, yMax_=crGraphs.INVALID_VALUE, yTicksStep_ = 0, yTicksStepMultiplier_ = 1,
                    filePath_ = "", renderFigure_=True, figure_=None, subPlot_=111,
                    ):
    """
    Create bars for groups next to each other, grouped by group labels. [!work in progress]
    """

    #-- test and pre-set data

    numOfDataDimensions = crData.getNumberOfListDimensions(data_);
    if (numOfDataDimensions < 2 or numOfDataDimensions > 3):
        raise ValueError("The data_ parameter must be a 2D or a 3D list.")

    if (len(xTickLabels_) > 0):
      crHelpers.checkListHasAtLeastLengthOfList(xTickLabels_, data_[0]);

    if (len(legendLabels_) > 0):
        crHelpers.checkListHasAtLeastLengthOfList(legendLabels_, data_);


    #-- apply styles
    legendFontSize_ = crGraphs.getStyle().legendFontSize;
    size_ = crGraphs.getStyle().figureSize;
    legendPosition_ = crGraphs.getStyle().legendPosition;
    colors_ = crGraphs.getStyle().colors;
    numOfLegendColumns_ = crGraphs.getStyle().numOfLegendColumns;
    labelFontSize_ = crGraphs.getStyle().labelFontSize;
    tickFontSize_ = crGraphs.getStyle().tickFontSize;


    xLocations = numpy.arange(len(data_[0]))  # the x locations for the groups
    fig, ax = crGraphs.createFigure(size_, title_, figure_, subPlot_, xLabel_, yLabel_);

    if (len(xTickLabels_) == 0):
       xTickLabels_ = [];
       xTickFrequency = max(1,math.floor(len(xLocations)/(size_[1]/0.5)));
       #-- make sure the tick frequency is a nice round number
       allowedFrequences = [1,5,10,25,50,100,250,500,1000];
       for f in allowedFrequences:
          if (xTickFrequency < f):
             xTickFrequency = f;
             break;
       #-- create numbered ticks of proper frequency, also marking the last tick value
       for i in range(len(xLocations)):
          #--  each i ticks               that are not too close to the end         plus the last tick
          if ((i%xTickFrequency == 0 and i < len(xLocations) - xTickFrequency/2 ) or (i == len(xLocations)-1)):
             xTickLabels_.append(i);
          else:
             xTickLabels_.append('');

    plt.xticks(xLocations, size=labelFontSize_);
    ax.set_xticklabels(xTickLabels_, size=tickFontSize_);


    #-- plot bars next to each other
    plots = [];
    maxVal = -9999999;
    minVal = 9999999;
    for i in range(len(data_)):
        #-- pick a color
        colorCode = 'b';
        if (len(colors_) > i):
            colorCode = colors_[i];

        if (numOfDataDimensions == 3):
            medians = [];
            for q in range(len(data_[i])):
                medians.append(numpy.median(data_[i][q]));

                for r in range(len(data_[i][q])):
                    if (data_[i][q][r] > maxVal):
                        maxVal = data_[i][q][r];
                    if (data_[i][q][r] < minVal):
                        minVal = data_[i][q][r]

            if (showConfidenceIntervals_):
                dataDof = [(len(data_[i][q])-1) for q in range(len(medians))]; #degrees of freedom is sample size -1
                dataStd = [numpy.std(data_[i][q]) for q in range(len(medians))];
                plot = ax.bar(xLocations+i*barWidth_, medians, barWidth_, color=colorCode, yerr=scipy.stats.t.ppf(0.95, dataDof)*dataStd); #yerr=stdData_[i]
                plots.append(plot);

            else:
                plot = ax.bar(xLocations+i*barWidth_, medians, barWidth_, color=colorCode, yerr=scipy.stats.t.ppf(0.95, dataDof)*dataStd); #yerr=stdData_[i]
                plots.append(plot);

        else:
            if (maxVal < crData.getMaxValueInAList(data_[i])):
                maxVal = crData.getMaxValueInAList(data_[i]);
            if (minVal > crData.getMinValueInAList(data_[i])):
                minVal = crData.getMinValueInAList(data_[i]);
            plot = ax.bar(xLocations+i*barWidth_, data_[i], barWidth_, color=colorCode); #yerr=stdData_[i]
            plots.append(plot);

    #-- adjust x axis and y axis limits:
    #ax.set_xlim(xLocations[0] - barWidth_*0.2, xLocations[-1] + barWidth_*1.2);


    crGraphs.setFigureAxisLimits(ax, minVal, maxVal, xMin_, xMax_, yMin_, yMax_, yTicksStep_, yTicksStepMultiplier_);

    #-- setup legend
    #box = ax.get_position()
    #legendItems = [];
    #for g in range(len(plots)):
    #    legendItems.append(plots[g][0]);

    #-- show legend
    if (len(legendLabels_) > 0):
        def flip(items, ncol):
            return itertools.chain(*[items[i::ncol] for i in range(ncol)])

        legendItems = [];
        for g in range(len(plots)):
            legendItems.append(plots[g][0]);
        legend = ax.legend(flip(legendItems, numOfLegendColumns_), flip(legendLabels_,numOfLegendColumns_),loc=legendPosition_.value, ncol=numOfLegendColumns_, fancybox=True, framealpha=0.7)
        for t in legend.get_texts():
            if (type(legendFontSize_) == str):
                t.set_fontsize(legendFontSize_)
            else:
                font = math.QFont(t.font());
                font.setPointSize(legendFontSize_);
                t.setFont(font);


    #-- display / print, return:
    crGraphs.renderFigure(fig, filePath_, renderFigure_);
    return fig;